/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file accuracy_exact.hh
	\brief Accuracy::Exact class, and associated functions
*/




#ifndef __ACCURACY_EXACT_HH__
#define __ACCURACY_EXACT_HH__

#ifdef __cplusplus




#include "accuracy_approximate.hh"
#include "refcount.hh"
#include "helpers.hh"

#include <gmpxx.h>

#include <boost/assert.hpp>
#include <boost/static_assert.hpp>

#include <ostream>
#include <algorithm>
#include <limits>
#include <cmath>




namespace Accuracy {




namespace _Private {




//============================================================================
//    NullType struct
//============================================================================


struct NullType;




//============================================================================
//    GetExactType helper template
//============================================================================


template< typename t_Type >
struct GetExactType {

	typedef t_Type Type;
};


template< typename t_Type >
struct GetExactType< t_Type const > {

	typedef typename GetExactType< t_Type >::Type Type;
};


template< typename t_Type >
struct GetExactType< t_Type& > {

	typedef typename GetExactType< t_Type >::Type Type;
};




//============================================================================
//    OperationCache class
//============================================================================


struct OperationCacheBase : public virtual Refcounted< false > {

	inline OperationCacheBase();

	virtual ~OperationCacheBase() = 0;


	virtual Approximate const& GetApproximate() = 0;
	virtual mpq_class const& GetExact() = 0;
};


template< typename t_OperationType >
struct OperationCache : public OperationCacheBase {

	inline OperationCache( t_OperationType const& operation );

	virtual ~OperationCache();


	virtual Approximate const& GetApproximate();
	virtual mpq_class const& GetExact();


private:

	t_OperationType m_operation;

	Approximate m_approximate;
	bool m_haveApproximate;

	mpq_class const* m_pExact;
};




//============================================================================
//    Operation enum
//============================================================================


enum {

	// nullary operations
	OPERATION_VARIABLE,

	// unary operations
	OPERATION_IDENTITY,
	OPERATION_NEGATE,
	OPERATION_SQUARE,
	OPERATION_POW,
	OPERATION_ABS,

	// binary operations
	OPERATION_ADD,
	OPERATION_SUBTRACT,
	OPERATION_MULTIPLY,
	OPERATION_DIVIDE,
	OPERATION_MIN,
	OPERATION_MAX
};




//============================================================================
//    Operation template declaration
//============================================================================


template< int t_Operation, typename t_LeftType, typename t_RightType >
struct Operation;




//============================================================================
//    Operation< OPERATION_VARIABLE, ... > template definition
//============================================================================


template<>
struct Operation< OPERATION_VARIABLE, NullType, NullType > {

	inline Operation();

	template< int t_Operation, typename t_LeftType, typename t_RightType >
	inline Operation( Operation< t_Operation, t_LeftType, t_RightType > const& operation );

	inline Operation( Operation const& other );


	template< int t_Operation, typename t_LeftType, typename t_RightType >
	inline Operation const& operator=( Operation< t_Operation, t_LeftType, t_RightType > const& operation );

	inline Operation const& operator=( Operation const& other );


	inline Accuracy::Approximate const& GetApproximate() const;
	inline mpq_class const& GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	RefcountedPointer< OperationCacheBase > m_pOperationCache;
};




//============================================================================
//    Operation< OPERATION_IDENTITY, ... > template definition
//============================================================================


template< typename t_Type >
struct Operation< OPERATION_IDENTITY, t_Type, NullType > {

	inline Operation( t_Type const& exact );


	inline t_Type const& GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	t_Type const m_exact;
};


template<>
struct Operation< OPERATION_IDENTITY, mpq_class, NullType > {

	inline Operation( mpq_class const& exact );


	inline Approximate const GetApproximate() const;
	inline mpq_class const& GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	mpq_class const m_exact;
};




//============================================================================
//    Operation< OPERATION_NEGATE, ... > template definition
//============================================================================


template< typename t_Type >
struct Operation< OPERATION_NEGATE, t_Type, NullType > {

	inline Operation( t_Type const& value );


	inline Approximate const GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline t_Type const& operator-() const;


private:

	t_Type const m_value;
};




//============================================================================
//    Operation< OPERATION_SQUARE, ... > template definition
//============================================================================


template< typename t_Type >
struct Operation< OPERATION_SQUARE, t_Type, NullType > {

	inline Operation( t_Type const& value );


	inline Approximate const GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	t_Type const m_value;
};



//============================================================================
//    Operation< OPERATION_POW, ... > template definition
//============================================================================


template< typename t_Type >
struct Operation< OPERATION_POW, t_Type, NullType > {

	inline Operation( t_Type const& value, int const power );


	inline Approximate const GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	t_Type const m_value;
	int const m_power;
};




//============================================================================
//    Operation< OPERATION_ABS, ... > template definition
//============================================================================


template< typename t_Type >
struct Operation< OPERATION_ABS, t_Type, NullType > {

	inline Operation( t_Type const& value );


	inline Approximate const GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	t_Type const m_value;
};




//============================================================================
//    Operation< OPERATION_ADD, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Operation< OPERATION_ADD, t_LeftType, t_RightType > {

	inline Operation( t_LeftType const& left, t_RightType const& right );


	inline Approximate const GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Operation< OPERATION_SUBTRACT, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Operation< OPERATION_SUBTRACT, t_LeftType, t_RightType > {

	inline Operation( t_LeftType const& left, t_RightType const& right );


	inline Approximate const GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Operation< OPERATION_MULTIPLY, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Operation< OPERATION_MULTIPLY, t_LeftType, t_RightType > {

	inline Operation( t_LeftType const& left, t_RightType const& right );


	inline Approximate const GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Operation< OPERATION_DIVIDE, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Operation< OPERATION_DIVIDE, t_LeftType, t_RightType > {

	inline Operation( t_LeftType const& left, t_RightType const& right );


	inline Approximate const GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Operation< OPERATION_MIN, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Operation< OPERATION_MIN, t_LeftType, t_RightType > {

	inline Operation( t_LeftType const& left, t_RightType const& right );


	inline Approximate const GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Operation< OPERATION_MAX, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Operation< OPERATION_MAX, t_LeftType, t_RightType > {

	inline Operation( t_LeftType const& left, t_RightType const& right );


	inline Approximate const GetApproximate() const;
	inline mpq_class const GetExact() const;


	inline Operation< OPERATION_NEGATE, Operation, NullType > const operator-() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    PredicateCache class
//============================================================================


struct PredicateCacheBase : public virtual Refcounted< false > {

	inline PredicateCacheBase();

	virtual ~PredicateCacheBase() = 0;


	virtual int const GetApproximate() = 0;
	virtual bool const GetExact() = 0;
};


template< typename t_PredicateType >
struct PredicateCache : public PredicateCacheBase {

	inline PredicateCache( t_PredicateType const& predicate );

	virtual ~PredicateCache();


	virtual int const GetApproximate();
	virtual bool const GetExact();


private:

	t_PredicateType m_predicate;

	int m_approximate;
	bool m_haveApproximate;

	bool m_exact;
	bool m_haveExact;
};




//============================================================================
//    Predicate enum
//============================================================================


enum {

	// nullary operations
	PREDICATE_VARIABLE,

	// unary operations
	PREDICATE_IDENTITY,
	PREDICATE_NOT,

	// binary operations
	PREDICATE_AND,
	PREDICATE_OR,
	PREDICATE_L,
	PREDICATE_LE,
	PREDICATE_G,
	PREDICATE_GE,
	PREDICATE_E,
	PREDICATE_NE
};




//============================================================================
//    Predicate template declaration
//============================================================================


template< int t_Predicate, typename t_LeftType, typename t_RightType >
struct Predicate;




//============================================================================
//    Predicate< PREDICATE_VARIABLE, ... > template definition
//============================================================================


template<>
struct Predicate< PREDICATE_VARIABLE, NullType, NullType > {

	inline Predicate();

	template< int t_Predicate, typename t_LeftType, typename t_RightType >
	inline Predicate( Predicate< t_Predicate, t_LeftType, t_RightType > const& predicate );

	inline Predicate( Predicate const& other );


	template< int t_Predicate, typename t_LeftType, typename t_RightType >
	inline Predicate const& operator=( Predicate< t_Predicate, t_LeftType, t_RightType > const& predicate );

	inline Predicate const& operator=( Predicate const& other );


	inline int const GetApproximate() const;    // -1 = false, 0 = unknown, 1 = true
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline Predicate< PREDICATE_NOT, Predicate, NullType > const operator!() const;


private:

	RefcountedPointer< PredicateCacheBase > m_pPredicateCache;
};




//============================================================================
//    Predicate< PREDICATE_IDENTITY, ... > helper template
//============================================================================


template<>
struct Predicate< PREDICATE_IDENTITY, NullType, NullType > {

public:

	inline Predicate( bool const value );


	inline int const GetApproximate() const;    // -1 = false, 0 = unknown, 1 = true
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline Predicate const operator!() const;


private:

	bool const m_value;
};




//============================================================================
//    Predicate< PREDICATE_NOT, ... > helper template
//============================================================================


template< typename t_Type >
struct Predicate< PREDICATE_NOT, t_Type, NullType > {

public:

	inline Predicate( t_Type const& value );


	inline int const GetApproximate() const;    // -1 = false, 0 = unknown, 1 = true
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline t_Type const& operator!() const;


private:

	t_Type const m_value;
};




//============================================================================
//    Predicate< PREDICATE_AND, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Predicate< PREDICATE_AND, t_LeftType, t_RightType > {

	inline Predicate( t_LeftType const& left, t_RightType const& right );


	inline int const GetApproximate() const;
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline Predicate< PREDICATE_NOT, Predicate, NullType > const operator!() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Predicate< PREDICATE_OR, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Predicate< PREDICATE_OR, t_LeftType, t_RightType > {

	inline Predicate( t_LeftType const& left, t_RightType const& right );


	inline int const GetApproximate() const;
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline Predicate< PREDICATE_NOT, Predicate, NullType > const operator!() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Predicate< PREDICATE_L, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Predicate< PREDICATE_L, t_LeftType, t_RightType > {

	inline Predicate( t_LeftType const& left, t_RightType const& right );


	inline int const GetApproximate() const;
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline Predicate< PREDICATE_GE, t_LeftType, t_RightType > const operator!() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Predicate< PREDICATE_LE, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Predicate< PREDICATE_LE, t_LeftType, t_RightType > {

	inline Predicate( t_LeftType const& left, t_RightType const& right );


	inline int const GetApproximate() const;
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline Predicate< PREDICATE_G, t_LeftType, t_RightType > const operator!() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Predicate< PREDICATE_G, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Predicate< PREDICATE_G, t_LeftType, t_RightType > {

	inline Predicate( t_LeftType const& left, t_RightType const& right );


	inline int const GetApproximate() const;
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline Predicate< PREDICATE_LE, t_LeftType, t_RightType > const operator!() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Predicate< PREDICATE_GE, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Predicate< PREDICATE_GE, t_LeftType, t_RightType > {

	inline Predicate( t_LeftType const& left, t_RightType const& right );


	inline int const GetApproximate() const;
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline Predicate< PREDICATE_L, t_LeftType, t_RightType > const operator!() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Predicate< PREDICATE_E, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Predicate< PREDICATE_E, t_LeftType, t_RightType > {

	inline Predicate( t_LeftType const& left, t_RightType const& right );


	inline int const GetApproximate() const;
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline Predicate< PREDICATE_NE, t_LeftType, t_RightType > const operator!() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    Predicate< PREDICATE_NE, ... > helper template
//============================================================================


template< typename t_LeftType, typename t_RightType >
struct Predicate< PREDICATE_NE, t_LeftType, t_RightType > {

	inline Predicate( t_LeftType const& left, t_RightType const& right );


	inline int const GetApproximate() const;
	inline bool const GetExact() const;

	inline operator bool const() const;


	inline Predicate< PREDICATE_E, t_LeftType, t_RightType > const operator!() const;


private:

	t_LeftType const m_left;
	t_RightType const m_right;
};




//============================================================================
//    OperationCache methods
//============================================================================


OperationCacheBase::OperationCacheBase() {
}


template< typename t_OperationType >
OperationCache< t_OperationType >::OperationCache( t_OperationType const& operation ) :
	m_operation( operation ),
	m_haveApproximate( false ),
	m_pExact( NULL )
{
}


template< typename t_OperationType >
OperationCache< t_OperationType >::~OperationCache() {

	if ( m_pExact != NULL )
		delete m_pExact;
}


template< typename t_OperationType >
Approximate const& OperationCache< t_OperationType >::GetApproximate() {

	if ( ! m_haveApproximate ) {

		m_approximate = Approximate( m_operation.GetApproximate() );
		m_haveApproximate = true;

		BOOST_ASSERT( m_pExact == NULL );
	}

	return m_approximate;
}


template< typename t_OperationType >
mpq_class const& OperationCache< t_OperationType >::GetExact() {

	if ( m_pExact == NULL ) {

		m_pExact = new mpq_class( m_operation.GetExact() );

		m_approximate = Approximate( m_pExact->get_d() );
		m_haveApproximate = true;
	}

	return *m_pExact;
}




//============================================================================
//    Operation< OPERATION_VARIABLE, ... > methods
//============================================================================


Operation< OPERATION_VARIABLE, NullType, NullType >::Operation() {
}


template< int t_Operation, typename t_LeftType, typename t_RightType >
Operation< OPERATION_VARIABLE, NullType, NullType >::Operation( Operation< t_Operation, t_LeftType, t_RightType > const& operation ) :
	m_pOperationCache( new OperationCache< Operation< t_Operation, t_LeftType, t_RightType > >( operation ) )
{
}


Operation< OPERATION_VARIABLE, NullType, NullType >::Operation( Operation const& other ) :
	m_pOperationCache( other.m_pOperationCache )
{
}


template< int t_Operation, typename t_LeftType, typename t_RightType >
Operation< OPERATION_VARIABLE, NullType, NullType > const& Operation< OPERATION_VARIABLE, NullType, NullType >::operator=( Operation< t_Operation, t_LeftType, t_RightType > const& operation ) {

	m_pOperationCache = new OperationCache< Operation< t_Operation, t_LeftType, t_RightType > >( operation );
	return *this;
}


Operation< OPERATION_VARIABLE, NullType, NullType > const& Operation< OPERATION_VARIABLE, NullType, NullType >::operator=( Operation const& other ) {

	m_pOperationCache = other.m_pOperationCache;
	return *this;
}


Approximate const& Operation< OPERATION_VARIABLE, NullType, NullType >::GetApproximate() const {

	BOOST_ASSERT( m_pOperationCache.Get() != NULL );
	return m_pOperationCache->GetApproximate();
}


mpq_class const& Operation< OPERATION_VARIABLE, NullType, NullType >::GetExact() const {

	BOOST_ASSERT( m_pOperationCache.Get() != NULL );
	return m_pOperationCache->GetExact();
}


Operation< OPERATION_NEGATE, Operation< OPERATION_VARIABLE, NullType, NullType >, NullType > const Operation< OPERATION_VARIABLE, NullType, NullType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_VARIABLE, NullType, NullType >, NullType >( *this );
}




//============================================================================
//    Operation< OPERATION_IDENTITY, ... > methods
//============================================================================


template< typename t_Type >
Operation< OPERATION_IDENTITY, t_Type, NullType >::Operation( t_Type const& exact ) : m_exact( exact ) {
}


template< typename t_Type >
t_Type const& Operation< OPERATION_IDENTITY, t_Type, NullType >::GetApproximate() const {

	return m_exact;
}


template< typename t_Type >
mpq_class const Operation< OPERATION_IDENTITY, t_Type, NullType >::GetExact() const {

	return mpq_class( m_exact );
}


template< typename t_Type >
Operation< OPERATION_NEGATE, Operation< OPERATION_IDENTITY, t_Type, NullType >, NullType > const Operation< OPERATION_IDENTITY, t_Type, NullType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_IDENTITY, t_Type, NullType >, NullType >( *this );
}


Operation< OPERATION_IDENTITY, mpq_class, NullType >::Operation( mpq_class const& exact ) : m_exact( exact ) {
}


Approximate const Operation< OPERATION_IDENTITY, mpq_class, NullType >::GetApproximate() const {

	return Approximate( m_exact.get_d() );
}


mpq_class const& Operation< OPERATION_IDENTITY, mpq_class, NullType >::GetExact() const {

	return m_exact;
}


Operation< OPERATION_NEGATE, Operation< OPERATION_IDENTITY, mpq_class, NullType >, NullType > const Operation< OPERATION_IDENTITY, mpq_class, NullType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_IDENTITY, mpq_class, NullType >, NullType >( *this );
}




//============================================================================
//    Operation< OPERATION_NEGATE, ... > methods
//============================================================================


template< typename t_Type >
Operation< OPERATION_NEGATE, t_Type, NullType >::Operation( t_Type const& value ) :
	m_value( value )
{
}


template< typename t_Type >
Approximate const Operation< OPERATION_NEGATE, t_Type, NullType >::GetApproximate() const {

	return -Approximate( m_value.GetApproximate() );
}


template< typename t_Type >
mpq_class const Operation< OPERATION_NEGATE, t_Type, NullType >::GetExact() const {

	return( -m_value.GetExact() );
}


template< typename t_Type >
t_Type const& Operation< OPERATION_NEGATE, t_Type, NullType >::operator-() const {

	return m_value;
}




//============================================================================
//    Operation< OPERATION_SQUARE, ... > methods
//============================================================================


template< typename t_Type >
Operation< OPERATION_SQUARE, t_Type, NullType >::Operation( t_Type const& value ) :
	m_value( value )
{
}


template< typename t_Type >
Approximate const Operation< OPERATION_SQUARE, t_Type, NullType >::GetApproximate() const {

	return ::Square( Approximate( m_value.GetApproximate() ) );
}


template< typename t_Type >
mpq_class const Operation< OPERATION_SQUARE, t_Type, NullType >::GetExact() const {

	return ::Square( m_value.GetExact() );
}


template< typename t_Type >
Operation< OPERATION_NEGATE, Operation< OPERATION_SQUARE, t_Type, NullType >, NullType > const Operation< OPERATION_SQUARE, t_Type, NullType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_SQUARE, t_Type, NullType >, NullType >( *this );
}




//============================================================================
//    Operation< OPERATION_POW, ... > methods
//============================================================================


template< typename t_Type >
Operation< OPERATION_POW, t_Type, NullType >::Operation( t_Type const& value, int const power ) :
	m_value( value ),
	m_power( power )
{
}


template< typename t_Type >
Approximate const Operation< OPERATION_POW, t_Type, NullType >::GetApproximate() const {

	return ::Pow( Approximate( m_value.GetApproximate() ), m_power );
}


template< typename t_Type >
mpq_class const Operation< OPERATION_POW, t_Type, NullType >::GetExact() const {

	return ::Pow( m_value.GetExact(), m_power );
}


template< typename t_Type >
Operation< OPERATION_NEGATE, Operation< OPERATION_POW, t_Type, NullType >, NullType > const Operation< OPERATION_POW, t_Type, NullType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_POW, t_Type, NullType >, NullType >( *this );
}




//============================================================================
//    Operation< OPERATION_ABS, ... > methods
//============================================================================


template< typename t_Type >
Operation< OPERATION_ABS, t_Type, NullType >::Operation( t_Type const& value ) :
	m_value( value )
{
}


template< typename t_Type >
Approximate const Operation< OPERATION_ABS, t_Type, NullType >::GetApproximate() const {

	return ::std::abs( Approximate( m_value.GetApproximate() ) );
}


template< typename t_Type >
mpq_class const Operation< OPERATION_ABS, t_Type, NullType >::GetExact() const {

	return ::std::abs( m_value.GetExact() );
}


template< typename t_Type >
Operation< OPERATION_NEGATE, Operation< OPERATION_ABS, t_Type, NullType >, NullType > const Operation< OPERATION_ABS, t_Type, NullType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_ABS, t_Type, NullType >, NullType >( *this );
}




//============================================================================
//    Operation< OPERATION_ADD, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_ADD, t_LeftType, t_RightType >::Operation( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
Approximate const Operation< OPERATION_ADD, t_LeftType, t_RightType >::GetApproximate() const {

	return( Approximate( m_left.GetApproximate() ) + Approximate( m_right.GetApproximate() ) );
}


template< typename t_LeftType, typename t_RightType >
mpq_class const Operation< OPERATION_ADD, t_LeftType, t_RightType >::GetExact() const {

	return( m_left.GetExact() + m_right.GetExact() );
}


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_NEGATE, Operation< OPERATION_ADD, t_LeftType, t_RightType >, NullType > const Operation< OPERATION_ADD, t_LeftType, t_RightType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_ADD, t_LeftType, t_RightType >, NullType >( *this );
}




//============================================================================
//    Operation< OPERATION_SUBTRACT, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_SUBTRACT, t_LeftType, t_RightType >::Operation( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
Approximate const Operation< OPERATION_SUBTRACT, t_LeftType, t_RightType >::GetApproximate() const {

	return( Approximate( m_left.GetApproximate() ) - Approximate( m_right.GetApproximate() ) );
}


template< typename t_LeftType, typename t_RightType >
mpq_class const Operation< OPERATION_SUBTRACT, t_LeftType, t_RightType >::GetExact() const {

	return( m_left.GetExact() - m_right.GetExact() );
}


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_NEGATE, Operation< OPERATION_SUBTRACT, t_LeftType, t_RightType >, NullType > const Operation< OPERATION_SUBTRACT, t_LeftType, t_RightType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_SUBTRACT, t_LeftType, t_RightType >, NullType >( *this );
}




//============================================================================
//    Operation< OPERATION_MULTIPLY, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_MULTIPLY, t_LeftType, t_RightType >::Operation( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
Approximate const Operation< OPERATION_MULTIPLY, t_LeftType, t_RightType >::GetApproximate() const {

	return( Approximate( m_left.GetApproximate() ) * Approximate( m_right.GetApproximate() ) );
}


template< typename t_LeftType, typename t_RightType >
mpq_class const Operation< OPERATION_MULTIPLY, t_LeftType, t_RightType >::GetExact() const {

	return( m_left.GetExact() * m_right.GetExact() );
}


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_NEGATE, Operation< OPERATION_MULTIPLY, t_LeftType, t_RightType >, NullType > const Operation< OPERATION_MULTIPLY, t_LeftType, t_RightType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_MULTIPLY, t_LeftType, t_RightType >, NullType >( *this );
}




//============================================================================
//    Operation< OPERATION_DIVIDE, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_DIVIDE, t_LeftType, t_RightType >::Operation( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
Approximate const Operation< OPERATION_DIVIDE, t_LeftType, t_RightType >::GetApproximate() const {

	return( Approximate( m_left.GetApproximate() ) / Approximate( m_right.GetApproximate() ) );
}


template< typename t_LeftType, typename t_RightType >
mpq_class const Operation< OPERATION_DIVIDE, t_LeftType, t_RightType >::GetExact() const {

	return( m_left.GetExact() / m_right.GetExact() );
}


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_NEGATE, Operation< OPERATION_DIVIDE, t_LeftType, t_RightType >, NullType > const Operation< OPERATION_DIVIDE, t_LeftType, t_RightType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_DIVIDE, t_LeftType, t_RightType >, NullType >( *this );
}




//============================================================================
//    Operation< OPERATION_MIN, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_MIN, t_LeftType, t_RightType >::Operation( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
Approximate const Operation< OPERATION_MIN, t_LeftType, t_RightType >::GetApproximate() const {

	return ::std::min( Approximate( m_left.GetApproximate() ), Approximate( m_right.GetApproximate() ) );
}


template< typename t_LeftType, typename t_RightType >
mpq_class const Operation< OPERATION_MIN, t_LeftType, t_RightType >::GetExact() const {

	return ::std::min( m_left.GetExact(), m_right.GetExact() );
}


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_NEGATE, Operation< OPERATION_MIN, t_LeftType, t_RightType >, NullType > const Operation< OPERATION_MIN, t_LeftType, t_RightType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_MIN, t_LeftType, t_RightType >, NullType >( *this );
}




//============================================================================
//    Operation< OPERATION_MAX, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_MAX, t_LeftType, t_RightType >::Operation( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
Approximate const Operation< OPERATION_MAX, t_LeftType, t_RightType >::GetApproximate() const {

	return ::std::max( Approximate( m_left.GetApproximate() ), Approximate( m_right.GetApproximate() ) );
}


template< typename t_LeftType, typename t_RightType >
mpq_class const Operation< OPERATION_MAX, t_LeftType, t_RightType >::GetExact() const {

	return ::std::max( m_left.GetExact(), m_right.GetExact() );
}


template< typename t_LeftType, typename t_RightType >
Operation< OPERATION_NEGATE, Operation< OPERATION_MAX, t_LeftType, t_RightType >, NullType > const Operation< OPERATION_MAX, t_LeftType, t_RightType >::operator-() const {

	return Operation< OPERATION_NEGATE, Operation< OPERATION_MAX, t_LeftType, t_RightType >, NullType >( *this );
}




//============================================================================
//    PredicateCache methods
//============================================================================


PredicateCacheBase::PredicateCacheBase() {
}


template< typename t_PredicateType >
PredicateCache< t_PredicateType >::PredicateCache( t_PredicateType const& predicate ) :
	m_predicate( predicate ),
	m_haveApproximate( false ),
	m_haveExact( false )
{
}


template< typename t_PredicateType >
PredicateCache< t_PredicateType >::~PredicateCache() {
}


template< typename t_PredicateType >
int const PredicateCache< t_PredicateType >::GetApproximate() {

	if ( ! m_haveApproximate ) {

		m_approximate = m_predicate.GetApproximate();
		m_haveApproximate = true;

		if ( m_approximate != 0 ) {

			m_exact = ( m_approximate > 0 );
			m_haveExact = true;
		}
	}

	return m_approximate;
}


template< typename t_PredicateType >
bool const PredicateCache< t_PredicateType >::GetExact() {

	if ( ! m_haveExact ) {

		m_exact = m_predicate.GetExact();
		m_haveExact = true;

		m_approximate = ( m_exact ? 1 : -1 );
		m_haveApproximate = true;
	}

	return m_exact;
}




//============================================================================
//    Predicate< PREDICATE_VARIABLE, ... > methods
//============================================================================


Predicate< PREDICATE_VARIABLE, NullType, NullType >::Predicate() {
}


template< int t_Predicate, typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_VARIABLE, NullType, NullType >::Predicate( Predicate< t_Predicate, t_LeftType, t_RightType > const& predicate ) :
	m_pPredicateCache( new PredicateCache< Predicate< t_Predicate, t_LeftType, t_RightType > >( predicate ) )
{
}


Predicate< PREDICATE_VARIABLE, NullType, NullType >::Predicate( Predicate const& other ) :
	m_pPredicateCache( other.m_pPredicateCache )
{
}


template< int t_Predicate, typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_VARIABLE, NullType, NullType > const& Predicate< PREDICATE_VARIABLE, NullType, NullType >::operator=( Predicate< t_Predicate, t_LeftType, t_RightType > const& predicate ) {

	m_pPredicateCache = new PredicateCache< Predicate< t_Predicate, t_LeftType, t_RightType > >( predicate );
	return *this;
}


Predicate< PREDICATE_VARIABLE, NullType, NullType > const& Predicate< PREDICATE_VARIABLE, NullType, NullType >::operator=( Predicate const& other ) {

	m_pPredicateCache = other.m_pPredicateCache;
	return *this;
}


int const Predicate< PREDICATE_VARIABLE, NullType, NullType >::GetApproximate() const {

	BOOST_ASSERT( m_pPredicateCache.Get() != NULL );
	return m_pPredicateCache->GetApproximate();
}


bool const Predicate< PREDICATE_VARIABLE, NullType, NullType >::GetExact() const {

	BOOST_ASSERT( m_pPredicateCache.Get() != NULL );
	return m_pPredicateCache->GetExact();
}


Predicate< PREDICATE_VARIABLE, NullType, NullType >::operator bool const() const {

	return GetExact();
}


Predicate< PREDICATE_NOT, Predicate< PREDICATE_VARIABLE, NullType, NullType >, NullType > const Predicate< PREDICATE_VARIABLE, NullType, NullType >::operator!() const {

	return Predicate< PREDICATE_NOT, Predicate< PREDICATE_VARIABLE, NullType, NullType >, NullType >( *this );
}




//============================================================================
//    Predicate< PREDICATE_IDENTITY, ... > methods
//============================================================================


Predicate< PREDICATE_IDENTITY, NullType, NullType >::Predicate( bool const value ) :
	m_value( value )
{
}


int const Predicate< PREDICATE_IDENTITY, NullType, NullType >::GetApproximate() const {

	return( m_value ? 1 : -1 );
}


bool const Predicate< PREDICATE_IDENTITY, NullType, NullType >::GetExact() const {

	return m_value;
}


Predicate< PREDICATE_IDENTITY, NullType, NullType >::operator bool const() const {

	return GetExact();
}


Predicate< PREDICATE_IDENTITY, NullType, NullType > const Predicate< PREDICATE_IDENTITY, NullType, NullType >::operator!() const {

	return Predicate( ! m_value );
}




//============================================================================
//    Predicate< PREDICATE_NOT, ... > methods
//============================================================================


template< typename t_Type >
Predicate< PREDICATE_NOT, t_Type, NullType >::Predicate( t_Type const& value ) :
	m_value( value )
{
}


template< typename t_Type >
int const Predicate< PREDICATE_NOT, t_Type, NullType >::GetApproximate() const {

	return -m_value.GetApproximate();
}


template< typename t_Type >
bool const Predicate< PREDICATE_NOT, t_Type, NullType >::GetExact() const {

	return( ! m_value.GetExact() );
}


template< typename t_Type >
Predicate< PREDICATE_NOT, t_Type, NullType >::operator bool const() const {

	return GetExact();
}


template< typename t_Type >
t_Type const& Predicate< PREDICATE_NOT, t_Type, NullType >::operator!() const {

	return m_value;
}




//============================================================================
//    Predicate< PREDICATE_AND, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_AND, t_LeftType, t_RightType >::Predicate( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
int const Predicate< PREDICATE_AND, t_LeftType, t_RightType >::GetApproximate() const {

	return ::std::min( m_left.GetApproximate(), m_right.GetApproximate() );
}


template< typename t_LeftType, typename t_RightType >
bool const Predicate< PREDICATE_AND, t_LeftType, t_RightType >::GetExact() const {

	switch( GetApproximate() ) {
		case 1: return true;
		case -1: return false;
		default: return( m_left.GetExact() && m_right.GetExact() );
	}
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_AND, t_LeftType, t_RightType >::operator bool const() const {

	return GetExact();
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_NOT, Predicate< PREDICATE_AND, t_LeftType, t_RightType >, NullType > const Predicate< PREDICATE_AND, t_LeftType, t_RightType >::operator!() const {

	return Predicate< PREDICATE_NOT, Predicate< PREDICATE_AND, t_LeftType, t_RightType >, NullType >( *this );
}




//============================================================================
//    Predicate< PREDICATE_OR, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_OR, t_LeftType, t_RightType >::Predicate( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
int const Predicate< PREDICATE_OR, t_LeftType, t_RightType >::GetApproximate() const {

	return ::std::max( m_left.GetApproximate(), m_right.GetApproximate() );
}


template< typename t_LeftType, typename t_RightType >
bool const Predicate< PREDICATE_OR, t_LeftType, t_RightType >::GetExact() const {

	switch( GetApproximate() ) {
		case 1: return true;
		case -1: return false;
		default: return( m_left.GetExact() || m_right.GetExact() );
	}
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_OR, t_LeftType, t_RightType >::operator bool const() const {

	return GetExact();
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_NOT, Predicate< PREDICATE_OR, t_LeftType, t_RightType >, NullType > const Predicate< PREDICATE_OR, t_LeftType, t_RightType >::operator!() const {

	return Predicate< PREDICATE_NOT, Predicate< PREDICATE_OR, t_LeftType, t_RightType >, NullType >( *this );
}




//============================================================================
//    Predicate< PREDICATE_L, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_L, t_LeftType, t_RightType >::Predicate( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
int const Predicate< PREDICATE_L, t_LeftType, t_RightType >::GetApproximate() const {

	return -Approximate::Compare( m_left.GetApproximate(), m_right.GetApproximate() );
}


template< typename t_LeftType, typename t_RightType >
bool const Predicate< PREDICATE_L, t_LeftType, t_RightType >::GetExact() const {

	switch( GetApproximate() ) {
		case 1: return true;
		case -1: return false;
		default: return( m_left.GetExact() < m_right.GetExact() );
	}
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_L, t_LeftType, t_RightType >::operator bool const() const {

	return GetExact();
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_GE, t_LeftType, t_RightType > const Predicate< PREDICATE_L, t_LeftType, t_RightType >::operator!() const {

	return Predicate< PREDICATE_GE, t_LeftType, t_RightType >( m_left, m_right );
}




//============================================================================
//    Predicate< PREDICATE_LE, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_LE, t_LeftType, t_RightType >::Predicate( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
int const Predicate< PREDICATE_LE, t_LeftType, t_RightType >::GetApproximate() const {

	return -Approximate::Compare( m_left.GetApproximate(), m_right.GetApproximate() );
}


template< typename t_LeftType, typename t_RightType >
bool const Predicate< PREDICATE_LE, t_LeftType, t_RightType >::GetExact() const {

	switch( GetApproximate() ) {
		case 1: return true;
		case -1: return false;
		default: return( m_left.GetExact() <= m_right.GetExact() );
	}
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_LE, t_LeftType, t_RightType >::operator bool const() const {

	return GetExact();
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_G, t_LeftType, t_RightType > const Predicate< PREDICATE_LE, t_LeftType, t_RightType >::operator!() const {

	return Predicate< PREDICATE_G, t_LeftType, t_RightType >( m_left, m_right );
}




//============================================================================
//    Predicate< PREDICATE_G, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_G, t_LeftType, t_RightType >::Predicate( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
int const Predicate< PREDICATE_G, t_LeftType, t_RightType >::GetApproximate() const {

	return Approximate::Compare( m_left.GetApproximate(), m_right.GetApproximate() );
}


template< typename t_LeftType, typename t_RightType >
bool const Predicate< PREDICATE_G, t_LeftType, t_RightType >::GetExact() const {

	switch( GetApproximate() ) {
		case 1: return true;
		case -1: return false;
		default: return( m_left.GetExact() > m_right.GetExact() );
	}
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_G, t_LeftType, t_RightType >::operator bool const() const {

	return GetExact();
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_LE, t_LeftType, t_RightType > const Predicate< PREDICATE_G, t_LeftType, t_RightType >::operator!() const {

	return Predicate< PREDICATE_LE, t_LeftType, t_RightType >( m_left, m_right );
}




//============================================================================
//    Predicate< PREDICATE_GE, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_GE, t_LeftType, t_RightType >::Predicate( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
int const Predicate< PREDICATE_GE, t_LeftType, t_RightType >::GetApproximate() const {

	return Approximate::Compare( m_left.GetApproximate(), m_right.GetApproximate() );
}


template< typename t_LeftType, typename t_RightType >
bool const Predicate< PREDICATE_GE, t_LeftType, t_RightType >::GetExact() const {

	switch( GetApproximate() ) {
		case 1: return true;
		case -1: return false;
		default: return( m_left.GetExact() >= m_right.GetExact() );
	}
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_GE, t_LeftType, t_RightType >::operator bool const() const {

	return GetExact();
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_L, t_LeftType, t_RightType > const Predicate< PREDICATE_GE, t_LeftType, t_RightType >::operator!() const {

	return Predicate< PREDICATE_L, t_LeftType, t_RightType >( m_left, m_right );
}




//============================================================================
//    Predicate< PREDICATE_E, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_E, t_LeftType, t_RightType >::Predicate( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
int const Predicate< PREDICATE_E, t_LeftType, t_RightType >::GetApproximate() const {

	return( ( Approximate::Compare( m_left.GetApproximate(), m_right.GetApproximate() ) == 0 ) ? 0 : -1 );
}


template< typename t_LeftType, typename t_RightType >
bool const Predicate< PREDICATE_E, t_LeftType, t_RightType >::GetExact() const {

	switch( GetApproximate() ) {
		case -1: return false;
		default: return( m_left.GetExact() == m_right.GetExact() );
	}
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_E, t_LeftType, t_RightType >::operator bool const() const {

	return GetExact();
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_NE, t_LeftType, t_RightType > const Predicate< PREDICATE_E, t_LeftType, t_RightType >::operator!() const {

	return Predicate< PREDICATE_NE, t_LeftType, t_RightType >( m_left, m_right );
}




//============================================================================
//    Predicate< PREDICATE_NE, ... > methods
//============================================================================


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_NE, t_LeftType, t_RightType >::Predicate( t_LeftType const& left, t_RightType const& right ) :
	m_left( left ),
	m_right( right )
{
}


template< typename t_LeftType, typename t_RightType >
int const Predicate< PREDICATE_NE, t_LeftType, t_RightType >::GetApproximate() const {

	return( ( Approximate::Compare( m_left.GetApproximate(), m_right.GetApproximate() ) == 0 ) ? 0 : 1 );
}


template< typename t_LeftType, typename t_RightType >
bool const Predicate< PREDICATE_NE, t_LeftType, t_RightType >::GetExact() const {

	switch( GetApproximate() ) {
		case 1: return true;
		default: return( m_left.GetExact() != m_right.GetExact() );
	}
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_NE, t_LeftType, t_RightType >::operator bool const() const {

	return GetExact();
}


template< typename t_LeftType, typename t_RightType >
Predicate< PREDICATE_E, t_LeftType, t_RightType > const Predicate< PREDICATE_NE, t_LeftType, t_RightType >::operator!() const {

	return Predicate< PREDICATE_E, t_LeftType, t_RightType >( m_left, m_right );
}




}    // namespace _Private




//============================================================================
//    Exact and ExactPredicate classes
//============================================================================


typedef _Private::Operation< _Private::OPERATION_VARIABLE, _Private::NullType, _Private::NullType > Exact;


typedef _Private::Predicate< _Private::PREDICATE_VARIABLE, _Private::NullType, _Private::NullType > ExactPredicate;




//============================================================================
//    ExactCast helper functions
//============================================================================


template< typename t_Type >
inline _Private::Operation< _Private::OPERATION_IDENTITY, t_Type, _Private::NullType > const ExactCast( t_Type const& value ) {

	return _Private::Operation< _Private::OPERATION_IDENTITY, t_Type, _Private::NullType >( value );
}


template< int t_Operation, typename t_LeftType, typename t_RightType >
inline _Private::Operation< t_Operation, t_LeftType, t_RightType > const& ExactCast( _Private::Operation< t_Operation, t_LeftType, t_RightType > const& value ) {

	return value;
}


inline _Private::Predicate< _Private::PREDICATE_IDENTITY, _Private::NullType, _Private::NullType > const ExactCast( bool const value ) {

	return _Private::Predicate< _Private::PREDICATE_IDENTITY, _Private::NullType, _Private::NullType >( value );
}


template< int t_Predicate, typename t_LeftType, typename t_RightType >
inline _Private::Predicate< t_Predicate, t_LeftType, t_RightType > const& ExactCast( _Private::Predicate< t_Predicate, t_LeftType, t_RightType > const& value ) {

	return value;
}




}    // namespace Accuracy




//============================================================================
//    inline comparison functions 
//============================================================================


template< int t_Operation, typename t_LeftType, typename t_RightType >
inline int const Sign( Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > const& value ) {

	int result = 0;

	switch( value.GetApproximate().Sign() ) {

		case  1: result =  1; break;
		case -1: result = -1; break;

		case 0: {
			if ( value.GetExact() < 0 )
				result = -1;
			else if ( value.GetExact() > 0 )
				result = 1;
			break;
		}
	}

	return result;
}




//============================================================================
//    inline comparison operators
//============================================================================


template< int t_LeftPredicate, typename t_LeftLeftType, typename t_LeftRightType, int t_RightPredicate, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_AND, Accuracy::_Private::Predicate< t_LeftPredicate, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Predicate< t_RightPredicate, t_RightLeftType, t_RightRightType > > const operator&&(
	Accuracy::_Private::Predicate< t_LeftPredicate, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Predicate< t_RightPredicate, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_AND, Accuracy::_Private::Predicate< t_LeftPredicate, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Predicate< t_RightPredicate, t_RightLeftType, t_RightRightType > >( left, right );
}


template< int t_LeftPredicate, typename t_LeftLeftType, typename t_LeftRightType, int t_RightPredicate, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_OR, Accuracy::_Private::Predicate< t_LeftPredicate, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Predicate< t_RightPredicate, t_RightLeftType, t_RightRightType > > const operator||(
	Accuracy::_Private::Predicate< t_LeftPredicate, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Predicate< t_RightPredicate, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_OR, Accuracy::_Private::Predicate< t_LeftPredicate, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Predicate< t_RightPredicate, t_RightLeftType, t_RightRightType > >( left, right );
}


template< int t_LeftOperation, typename t_LeftLeftType, typename t_LeftRightType, int t_RightOperation, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_L, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > > const operator<(
	Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_L, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > >( left, right );
}


template< int t_LeftOperation, typename t_LeftLeftType, typename t_LeftRightType, int t_RightOperation, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_LE, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > > const operator<=(
	Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_LE, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > >( left, right );
}


template< int t_LeftOperation, typename t_LeftLeftType, typename t_LeftRightType, int t_RightOperation, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_G, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > > const operator>(
	Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_G, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > >( left, right );
}


template< int t_LeftOperation, typename t_LeftLeftType, typename t_LeftRightType, int t_RightOperation, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_GE, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > > const operator>=(
	Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_GE, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > >( left, right );
}


template< int t_LeftOperation, typename t_LeftLeftType, typename t_LeftRightType, int t_RightOperation, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_E, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > > const operator==(
	Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_E, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > >( left, right );
}


template< int t_LeftOperation, typename t_LeftLeftType, typename t_LeftRightType, int t_RightOperation, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_NE, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > > const operator!=(
	Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Predicate< Accuracy::_Private::PREDICATE_NE, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > >( left, right );
}




//============================================================================
//    inline operators
//============================================================================


template< int t_LeftOperation, typename t_LeftLeftType, typename t_LeftRightType, int t_RightOperation, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_ADD, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > > const operator+(
	Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_ADD, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > >( left, right );
}


template< int t_LeftOperation, typename t_LeftLeftType, typename t_LeftRightType, int t_RightOperation, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_SUBTRACT, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > > const operator-(
	Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_SUBTRACT, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > >( left, right );
}


template< int t_LeftOperation, typename t_LeftLeftType, typename t_LeftRightType, int t_RightOperation, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_MULTIPLY, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > > const operator*(
	Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_MULTIPLY, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > >( left, right );
}


template< int t_LeftOperation, typename t_LeftLeftType, typename t_LeftRightType, int t_RightOperation, typename t_RightLeftType, typename t_RightRightType >
inline Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_DIVIDE, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > > const operator/(
	Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType > const& left,
	Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > const& right
)
{
	return Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_DIVIDE, Accuracy::_Private::Operation< t_LeftOperation, t_LeftLeftType, t_LeftRightType >, Accuracy::_Private::Operation< t_RightOperation, t_RightLeftType, t_RightRightType > >( left, right );
}




//============================================================================
//    inline helpers
//============================================================================


template< int t_Operation, typename t_LeftType, typename t_RightType >
inline Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_SQUARE, Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType >, Accuracy::_Private::NullType > const Square(
	Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > const& value
)
{
	return Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_SQUARE, Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType >, Accuracy::_Private::NullType >( value );
}


template< int t_Operation, typename t_LeftType, typename t_RightType >
inline Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_POW, Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType >, Accuracy::_Private::NullType > const Pow(
	Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > const& value,
	int const power
)
{
	return Accuracy::_Private::Operation< Accuracy::_Private::OPERATION_POW, Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType >, Accuracy::_Private::NullType >( value, power );
}




//============================================================================
//    inline helpers
//============================================================================


namespace std {


template< int t_Operation, typename t_LeftType, typename t_RightType >
inline ::Accuracy::_Private::Operation< ::Accuracy::_Private::OPERATION_ABS, ::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType >, ::Accuracy::_Private::NullType > const AbsoluteValue(
	::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > const& value
)
{
	return ::Accuracy::_Private::Operation< ::Accuracy::_Private::OPERATION_ABS, ::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType >, ::Accuracy::_Private::NullType >( value );
}


template< int t_Operation, typename t_LeftType, typename t_RightType >
inline ::Accuracy::_Private::Operation< ::Accuracy::_Private::OPERATION_MIN, ::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType >, ::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > > const min(
	::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > const& left,
	::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > const& right
)
{
	return ::Accuracy::_Private::Operation< ::Accuracy::_Private::OPERATION_MIN, ::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType >, ::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > >( left, right );
}


template< int t_Operation, typename t_LeftType, typename t_RightType >
inline ::Accuracy::_Private::Operation< ::Accuracy::_Private::OPERATION_MAX, ::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType >, ::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > > const max(
	::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > const& left,
	::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > const& right
)
{
	return ::Accuracy::_Private::Operation< ::Accuracy::_Private::OPERATION_MAX, ::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType >, ::Accuracy::_Private::Operation< t_Operation, t_LeftType, t_RightType > >( left, right );
}


}    // namespace std




#endif    /* __cplusplus */

#endif    /* __ACCURACY_EXACT_HH__ */
