/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file choose.hh
	\brief Choose class
*/




#ifndef __CHOOSE_HH__
#define __CHOOSE_HH__

#ifdef __cplusplus




#include <boost/shared_array.hpp>
#include <boost/assert.hpp>

#include <ostream>
#include <algorithm>




//============================================================================
//    Choose class
//============================================================================


/**
	\brief Class for enumerating all fixed-sized subsets of an index set

	Best illustrated with an example. The following code:
		for ( Choose choose( 6, 4 ); ! choose.Done(); ++choose )
			std::cout << choose << std::endl;
	Will have the following output:
		{ 0, 1, 2, 3 }
		{ 0, 1, 2, 4 }
		{ 0, 1, 2, 5 }
		{ 0, 1, 3, 4 }
		{ 0, 1, 3, 5 }
		{ 0, 1, 4, 5 }
		{ 0, 2, 3, 4 }
		{ 0, 2, 3, 5 }
		{ 0, 2, 4, 5 }
		{ 0, 3, 4, 5 }
		{ 1, 2, 3, 4 }
		{ 1, 2, 3, 5 }
		{ 1, 2, 4, 5 }
		{ 1, 3, 4, 5 }
		{ 2, 3, 4, 5 }
*/
struct Choose {

	inline Choose( unsigned int const top, unsigned int const bottom );
	inline Choose( Choose const& other );


	inline Choose const& operator=( Choose const& other );


	inline bool const Done() const;
	inline unsigned int const GetTop() const;
	inline unsigned int const GetBottom() const;
	inline unsigned int const operator[]( unsigned int const index ) const;


	inline Choose const& operator++();
	inline Choose const operator++( int );


private:

	unsigned int m_top;
	unsigned int m_bottom;

	boost::shared_array< unsigned int > m_indices;
};




//============================================================================
//    Choose inline methods
//============================================================================


Choose::Choose( unsigned int const top, unsigned int const bottom ) :
	m_top(    top    ),
	m_bottom( bottom ),
	m_indices( new unsigned int[ m_bottom ] )
{
	BOOST_ASSERT( m_bottom > 0 );
	BOOST_ASSERT( m_top >= m_bottom );

	for ( unsigned int ii = 0; ii < m_bottom; ++ii )
		m_indices[ ii ] = ( m_bottom - 1 ) - ii;
}


Choose::Choose( Choose const& other ) :
	m_top(    other.m_top    ),
	m_bottom( other.m_bottom ),
	m_indices( new unsigned int[ m_bottom ] )
{
	BOOST_ASSERT( m_bottom > 0 );
	BOOST_ASSERT( m_top >= m_bottom );

	std::copy( other.m_indices.get(), other.m_indices.get() + m_bottom, m_indices.get() );
}


Choose const& Choose::operator=( Choose const& other ) {

	if ( &other != this ) {

		m_top    = other.m_top;
		m_bottom = other.m_bottom;
		m_indices = boost::shared_array< unsigned int >( new unsigned int[ m_bottom ] );

		BOOST_ASSERT( m_bottom > 0 );
		BOOST_ASSERT( m_top >= m_bottom );

		std::copy( other.m_indices.get(), other.m_indices.get() + m_bottom, m_indices.get() );
	}

	return *this;
}


bool const Choose::Done() const {

	return( m_indices[ 0 ] >= m_top );
}


unsigned int const Choose::GetTop() const {

	return m_top;
}


unsigned int const Choose::GetBottom() const {

	return m_bottom;
}


unsigned int const Choose::operator[]( unsigned int const index ) const {

	BOOST_ASSERT( index < m_bottom );
	return m_indices[ ( m_bottom - 1 ) - index ];
}


Choose const& Choose::operator++() {

	BOOST_ASSERT( ! Done() );

	for ( unsigned int ii = 0; ii < m_bottom; ++ii ) {

		++m_indices[ ii ];
		if ( m_indices[ ii ] + ii < m_top ) {

			for ( ; ii > 0; --ii )
				m_indices[ ii - 1 ] = m_indices[ ii ] + 1;
			break;
		}
	}

	return *this;
}


Choose const Choose::operator++( int ) {

	Choose result( *this );
	operator++();
	return result;
}




//============================================================================
//    Choose inline helpers
//============================================================================


inline std::ostream& operator<<( std::ostream& out, Choose const& choose ) {

	out << "{ " << choose[ 0 ];
	for ( unsigned int ii = 1; ii < choose.GetBottom(); ++ii )
		out << ", " << choose[ ii ];
	out << " }";

	return out;
}




#endif    /* __cplusplus */

#endif    /* __CHOOSE_HH__ */
