/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file wsp.hh
	\brief WSP class
*/




#ifndef __WSP_HH__
#define __WSP_HH__

#ifdef __cplusplus




#include "quadtree.hh"
#include "accuracy.hh"
#include "maybe_threaded.hh"
#include "refcount.hh"
#include "pool.hh"
#include "helpers.hh"

#include <boost/bind.hpp>
#include <boost/function.hpp>
#include <boost/array.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/assert.hpp>
#include <boost/static_assert.hpp>

#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include <stdexcept>
#include <cmath>




//============================================================================
//    WSP class
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads = 1, typename t_Type = double >
struct WSP {

	BOOST_STATIC_ASSERT( t_Dimension > 0 );
	BOOST_STATIC_ASSERT( t_Threads > 0 );


	enum { DIMENSION = t_Dimension };
	typedef t_Type Type;


private:

	struct PointData;

	typedef Quadtree< t_Dimension, t_Threads, t_Type, RefcountedPointer< PointData > > QT;


public:

	typedef typename QT::Handle Handle;


	inline WSP(
		double const rho,
		double const beta,
		unsigned int const maximumColors,
		unsigned int const nodeSize,
		double const delta = 1.0
	);


	static inline boost::array< t_Type, t_Dimension > const& GetPoint( Handle const& handle );
	static inline bool const IsSteiner( Handle const& handle );


	inline unsigned int const GetPoints() const;
	inline unsigned int const GetSteinerPoints() const;


	inline bool const EnumeratePoints(        boost::function1< bool const, Handle const& > const& enumerator ) const;    // enumerator is Handle -> bool, false iff enumeration should stop
	inline bool const EnumerateSteinerPoints( boost::function1< bool const, Handle const& > const& enumerator ) const;    // enumerator is Handle -> bool, false iff enumeration should stop


	inline bool const EnumerateRecentlyInsertedSteinerPoints( boost::function1< bool const, Handle const& > const& enumerator ) const;    // enumerator is Handle -> bool, false iff enumeration should stop
	inline bool const EnumerateRecentlyErasedSteinerPoints(   boost::function1< bool const, Handle const& > const& enumerator ) const;    // enumerator is Handle -> bool, false iff enumeration should stop


	inline Handle const FindNearestNeighbor( Handle const& handle ) const;

	inline bool const EnumerateVoronoiNeighbors(
		Handle const& handle,
		boost::function1< bool const, Handle const& > const& enumerator    // enumerator is Handle -> bool, false iff enumeration should stop
	) const;


	Handle const Insert( boost::array< t_Type, t_Dimension > const& vertex );
	void Erase( Handle const& handle );


	void Refine( unsigned int const threads = t_Threads );


private:

	struct Operation : public virtual Refcounted< ( t_Threads > 1 ) > {

		inline Operation(
			Handle const& targetHandle,
			Handle const& sourceHandle,
			int const rank = std::numeric_limits< int >::max() + 1,    // only matters for dispatches and insertions
			bool const color = false                                   // only matters for insertions
		);

		inline Operation( Operation const& other );


		inline Handle const& GetTargetHandle() const;
		inline Handle const& GetSourceHandle() const;

		inline int const GetRank() const;
		inline uint64_t const GetColor() const;
		inline uint64_t const GetColorCoordinate() const;

		inline bool const GetEnqueued() const;


		inline void SetEnqueued( bool const enqueued );


		// A *partial* order on operations, representing which operations are guaranteed to execute before (or after) others
		inline bool const operator<( Operation const& other ) const;
		inline bool const operator>=( Operation const& other ) const;
		inline bool const operator>( Operation const& other ) const;
		inline bool const operator<=( Operation const& other ) const;


	private:

		Handle m_targetHandle;
		Handle m_sourceHandle;

		int m_rank;
		uint64_t m_color;
		uint64_t m_colorCoordinate;

		bool volatile m_enqueued;
	};


	// A *total* order on operations, representing the order in which they should be stored in the queue
	struct CompareOperations {

		inline CompareOperations();


		inline bool const operator()(
			RefcountedPointer< Operation > const& pFirst,
			RefcountedPointer< Operation > const& pSecond
		) const;
	};


	struct Functor {

		inline Functor(
			WSP* const pWSP,
			RefcountedPointer< Operation > const& pOperation
		);


		inline bool const operator()( Handle const& handle ) const;


	private:

		WSP* m_pWSP;

		RefcountedPointer< Operation > m_pOperation;
	};


	struct InsertCallback {

		inline InsertCallback(
			WSP* const pWSP,
			RefcountedPointer< Operation > const& pOperation
		);


		inline void operator()( Handle const& handle ) const;


	private:

		WSP* m_pWSP;
		RefcountedPointer< Operation > m_pOperation;
	};


	struct InitializeNodeList : public QT::NodeList {

		inline InitializeNodeList( PointData* const pPointData );


		virtual void SplitNotification();
		virtual void MergeNotification();


	private:

		PointData* m_pPointData;
	};


	struct DispatchNodeList : public QT::NodeList {

		inline DispatchNodeList( PointData* const pPointData );


		virtual void InsertNotification();
		virtual void InsertSteinerNotification( unsigned int const thread );
		virtual void EraseNotification();
		virtual void EraseSteinerNotification( unsigned int const thread );


	private:

		PointData* m_pPointData;
	};


	struct InsertNodeList : public QT::NodeList {

		inline InsertNodeList( PointData* const pPointData );


		virtual void InsertNotification();
		virtual void InsertSteinerNotification( unsigned int const thread );
		virtual void EraseNotification();
		virtual void EraseSteinerNotification( unsigned int const thread );


	private:

		PointData* m_pPointData;
	};


	struct PointData : public virtual Refcounted< ( t_Threads > 1 ) > {

		inline PointData(
			WSP* const pWSP,
			Handle const& handle,
			RefcountedPointer< Operation > const& pOperation
		);


		inline bool const IsErased() const;


		inline void EraseSteiner( unsigned int const thread );

		inline void Cleanup( unsigned int const thread );


		inline WSP* const GetWSP() const;

		inline Handle const& GetHandle() const;

		inline bool const IsSteiner() const;

		inline RefcountedPointer< Operation > const& GetOperation() const;


		void PerformInitializeOperation( PoolList& poolList, unsigned int const thread, RefcountedPointer< Operation > const& pOperation );
		void PerformDispatchOperation(   PoolList& poolList, unsigned int const thread, RefcountedPointer< Operation > const& pOperation );
		void PerformInsertOperation(     PoolList& poolList, unsigned int const thread, RefcountedPointer< Operation > const& pOperation );


	private:

		inline void ScheduleInitializeOperation( unsigned int const thread, bool const force = false );
		inline void ScheduleDispatchOperation(   unsigned int const thread, bool const force = false );
		inline void ScheduleInsertOperation(     unsigned int const thread, bool const force = false );

		inline void UndoDispatchOperation( unsigned int const thread );
		inline void UndoInsertOperation( unsigned int const thread, RefcountedPointer< Operation > const& pInsertOperation );


		WSP* m_pWSP;

		Handle const m_handle;

		bool volatile m_erased;

		RefcountedPointer< Operation > m_pOperation;    // NULL for normal points, operation which created this point for Steiners

		InitializeNodeList m_initializeNodeList;
		DispatchNodeList m_dispatchNodeList;
		InsertNodeList m_insertNodeList;

		RefcountedPointer< Operation > m_pActiveInitializeOperation;
		RefcountedPointer< Operation > m_pActiveDispatchOperation;
		RefcountedPointer< Operation > m_pActiveInsertOperation;

		RefcountedPointer< Operation > m_pSelfInitializeOperation;
		RefcountedPointer< Operation > m_pSelfDispatchOperation;
		RefcountedPointer< Operation > m_pSelfInsertOperation;

		std::set< RefcountedPointer< Operation > > m_pDispatchOperations;
		std::set< RefcountedPointer< Operation > > m_pInsertOperations;

		RefcountedPointer< Operation > m_pReverseDispatchOperation;
		std::set< RefcountedPointer< Operation > > m_pReverseInsertOperations;


		friend struct WSP;
		friend struct InitializeNodeList;
		friend struct DispatchNodeList;
		friend struct InsertNodeList;
	};


	inline int const FindRank( Accuracy::Exact const& distanceSquared ) const;

	void Thread(
		typename MaybeThreaded< ( t_Threads > 1 ) >::Barrier* const pBarrier,
		unsigned int const thread
	);


	double m_rho;
	double m_logRankRho;
	double m_rankRho;
	double m_beta;
	double m_colorScale;
	unsigned int m_colors;

	typename std::multiset< RefcountedPointer< Operation >, CompareOperations > m_pInitializeOperations;
	typename std::multiset< RefcountedPointer< Operation >, CompareOperations > m_pDispatchOperations;
	typename std::multiset< RefcountedPointer< Operation >, CompareOperations > m_pInsertOperations;
	typename std::set< Handle > m_erasedSteinerHandles;

	typename MaybeThreaded< ( t_Threads > 1 ) >::Mutex m_mutex;

	RefcountedPointer< Operation > m_pActiveOperations[ t_Threads ];
	typename std::multiset< RefcountedPointer< Operation >, CompareOperations > m_pLocalDispatchOperations[ t_Threads ];
	typename std::multiset< RefcountedPointer< Operation >, CompareOperations > m_pLocalInsertOperations[   t_Threads ];

	int volatile m_rank;
	uint64_t volatile m_color;

	QT m_quadtree;
	typename std::set< Handle > m_lessRecentlyInsertedSteinerHandles;
	typename std::set< Handle > m_lessRecentlyErasedSteinerHandles;
	typename std::set< Handle > m_recentlyInsertedSteinerHandles;
	typename std::set< Handle > m_recentlyErasedSteinerHandles;
};




//============================================================================
//    WSP::Operation inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >::Operation::Operation(
	Handle const& targetHandle,
	Handle const& sourceHandle,
	int const rank,
	bool const color
) :
	m_targetHandle( targetHandle ),
	m_sourceHandle( sourceHandle ),
	m_rank( rank ),
	m_color( 0 ),
	m_colorCoordinate( 0 ),
	m_enqueued( false )
{
	if ( color ) {

		RefcountedPointer< PointData > const& pPointData( QT::GetClientData( m_targetHandle ).Get() );
		BOOST_ASSERT( pPointData.Get() != NULL );
		BOOST_ASSERT( pPointData->GetHandle() == m_targetHandle );

		WSP* const pWSP( pPointData->GetWSP() );
		BOOST_ASSERT( pWSP != NULL );

		boost::array< t_Type, t_Dimension > const& vertex( QT::GetPoint( m_targetHandle ) );

		double const denominator = pWSP->m_colorScale * std::pow( pWSP->m_rankRho, m_rank - 0.1 ) / std::sqrt( t_Dimension );

		for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

			uint64_t const ratio = vertex[ ii ] / denominator;

			m_color *= pWSP->m_colors;
			m_color += ratio % pWSP->m_colors;

			m_colorCoordinate = ( m_colorCoordinate * 9219741426499971445ull ) + 1;    // from TestU01
			m_colorCoordinate += ratio / pWSP->m_colors;
		}

		// make sure colors are always nonzero (so that dispatches come before insertions)
		m_color           |= ( 1ull << ( 8 * sizeof( m_color           ) - 1 ) );
		m_colorCoordinate |= ( 1ull << ( 8 * sizeof( m_colorCoordinate ) - 1 ) );
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >::Operation::Operation( Operation const& other ) :
	m_targetHandle( other.m_targetHandle ),
	m_sourceHandle( other.m_sourceHandle ),
	m_rank( other.m_rank ),
	m_color( other.m_color ),
	m_colorCoordinate( other.m_colorCoordinate ),
	m_enqueued( false )
{
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
typename WSP< t_Dimension, t_Threads, t_Type >::Handle const& WSP< t_Dimension, t_Threads, t_Type >::Operation::GetTargetHandle() const {

	return m_targetHandle;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
typename WSP< t_Dimension, t_Threads, t_Type >::Handle const& WSP< t_Dimension, t_Threads, t_Type >::Operation::GetSourceHandle() const {

	return m_sourceHandle;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
int const WSP< t_Dimension, t_Threads, t_Type >::Operation::GetRank() const {

	return m_rank;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
uint64_t const WSP< t_Dimension, t_Threads, t_Type >::Operation::GetColor() const {

	return m_color;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
uint64_t const WSP< t_Dimension, t_Threads, t_Type >::Operation::GetColorCoordinate() const {

	return m_colorCoordinate;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::Operation::GetEnqueued() const {

	return( __sync_bool_compare_and_swap( const_cast< bool volatile* >( &m_enqueued ), true, true ) );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::Operation::SetEnqueued( bool const enqueued ) {

	__sync_bool_compare_and_swap( &m_enqueued, ( ! enqueued ), enqueued );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::Operation::operator<( Operation const& other ) const {

	bool result = false;
	if ( m_rank < other.m_rank )
		result = true;
	else if ( m_rank == other.m_rank ) {

		if ( m_color < other.m_color )
			result = true;
		else if ( ( m_color == other.m_color ) && ( m_color != 0 ) && ( m_colorCoordinate == other.m_colorCoordinate ) ) {

			if ( m_targetHandle < other.m_targetHandle )
				result = true;
			else if ( m_targetHandle == other.m_targetHandle ) {

				if ( m_sourceHandle < other.m_sourceHandle )
					result = true;
			}
		}
	}
	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::Operation::operator>=( Operation const& other ) const {

	return( ! operator<( other ) );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::Operation::operator>( Operation const& other ) const {

	bool result = false;
	if ( m_rank > other.m_rank )
		result = true;
	else if ( m_rank == other.m_rank ) {

		if ( m_color > other.m_color )
			result = true;
		else if ( ( m_color == other.m_color ) && ( m_color != 0 ) && ( m_colorCoordinate == other.m_colorCoordinate ) ) {

			if ( m_targetHandle > other.m_targetHandle )
				result = true;
			else if ( m_targetHandle == other.m_targetHandle ) {

				if ( m_sourceHandle > other.m_sourceHandle )
					result = true;
			}
		}
	}
	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::Operation::operator<=( Operation const& other ) const {

	return( ! operator>( other ) );
}




//============================================================================
//    WSP::CompareOperations methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >::CompareOperations::CompareOperations() {
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::CompareOperations::operator()(
	RefcountedPointer< Operation > const& pFirst,
	RefcountedPointer< Operation > const& pSecond
) const
{
	bool result = false;
	if ( pSecond.Get() != NULL ) {

		if ( pFirst.Get() == NULL )
			result = true;
		else {

			if ( pFirst->GetRank() < pSecond->GetRank() )
				result = true;
			else if ( pFirst->GetRank() == pSecond->GetRank() ) {

				if ( pFirst->GetColor() < pSecond->GetColor() )
					result = true;
				else if ( pFirst->GetColor() == pSecond->GetColor() ) {

					if ( pFirst->GetColorCoordinate() < pSecond->GetColorCoordinate() )
						result = true;
					else if ( pFirst->GetColorCoordinate() == pSecond->GetColorCoordinate() ) {

						if ( pFirst->GetTargetHandle() < pSecond->GetTargetHandle() )
							result = true;
						else if ( pFirst->GetTargetHandle() == pSecond->GetTargetHandle() ) {

							if ( pFirst->GetSourceHandle() < pSecond->GetSourceHandle() )
								result = true;
						}
					}
				}
			}
		}
	}
	return result;
}




//============================================================================
//    WSP::Functor inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >::Functor::Functor(
	WSP* const pWSP,
	RefcountedPointer< Operation > const& pOperation
) :
	m_pWSP( pWSP ),
	m_pOperation( pOperation )
{
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::Functor::operator()( Handle const& handle ) const {

	bool result = false;

	RefcountedPointer< PointData > const& pPointData( QT::GetClientData( handle ).Get() );
	BOOST_ASSERT( pPointData.Get() != NULL );
	BOOST_ASSERT( pPointData->GetHandle() == handle );

	if ( ! pPointData->IsErased() ) {

		RefcountedPointer< Operation > const& pOperation = pPointData->GetOperation();
		result = ( ( m_pOperation.Get() != NULL ) && ( ( pOperation.Get() == NULL ) || ( *pOperation < *m_pOperation ) ) );
	}

	return result;
}




//============================================================================
//    WSP::InsertCallback inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >::InsertCallback::InsertCallback(
	WSP* const pWSP,
	RefcountedPointer< Operation > const& pOperation
) :
	m_pWSP( pWSP ),
	m_pOperation( pOperation )
{
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::InsertCallback::operator()( Handle const& handle ) const {

	RefcountedPointer< PointData > pPointData( new PointData( m_pWSP, handle, m_pOperation ) );
	QT::GetClientData( handle ).Set( pPointData );
}




//============================================================================
//    WSP::InitializeNodeList inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >::InitializeNodeList::InitializeNodeList( PointData* const pPointData ) : m_pPointData( pPointData ) {
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::InitializeNodeList::SplitNotification() {

	if ( m_pPointData->m_pActiveInitializeOperation.Get() != NULL )
		m_pPointData->ScheduleInitializeOperation( 0 );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::InitializeNodeList::MergeNotification() {

	if ( m_pPointData->m_pActiveInitializeOperation.Get() != NULL )
		m_pPointData->ScheduleInitializeOperation( 0 );
}




//============================================================================
//    WSP::DispatchNodeList inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >::DispatchNodeList::DispatchNodeList( PointData* const pPointData ) : m_pPointData( pPointData ) {
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::DispatchNodeList::InsertNotification() {

	if ( m_pPointData->m_pActiveDispatchOperation.Get() != NULL )
		m_pPointData->ScheduleDispatchOperation( 0 );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::DispatchNodeList::InsertSteinerNotification( unsigned int const thread ) {

	if ( ! m_pPointData->IsErased() ) {

		RefcountedPointer< Operation > const& pLeftOperation  = m_pPointData->m_pWSP->m_pActiveOperations[ thread ];
		RefcountedPointer< Operation > const& pRightOperation = m_pPointData->m_pActiveDispatchOperation;
		if ( ( pRightOperation.Get() != NULL ) && ( ( pLeftOperation.Get() == NULL ) || ( *pLeftOperation < *pRightOperation ) ) )
			m_pPointData->ScheduleDispatchOperation( thread );
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::DispatchNodeList::EraseNotification() {

	if ( m_pPointData->m_pActiveDispatchOperation.Get() != NULL )
		m_pPointData->ScheduleDispatchOperation( 0 );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::DispatchNodeList::EraseSteinerNotification( unsigned int const thread ) {

	if ( ! m_pPointData->IsErased() ) {

		RefcountedPointer< Operation > const& pLeftOperation  = m_pPointData->m_pWSP->m_pActiveOperations[ thread ];
		RefcountedPointer< Operation > const& pRightOperation = m_pPointData->m_pActiveDispatchOperation;
		if ( ( pRightOperation.Get() != NULL ) && ( ( pLeftOperation.Get() == NULL ) || ( *pLeftOperation < *pRightOperation ) ) )
			m_pPointData->ScheduleDispatchOperation( thread );
	}
}




//============================================================================
//    WSP::InsertNodeList inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >::InsertNodeList::InsertNodeList( PointData* const pPointData ) : m_pPointData( pPointData ) {
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::InsertNodeList::InsertNotification() {

	if ( m_pPointData->m_pActiveInsertOperation.Get() != NULL )
		m_pPointData->ScheduleInsertOperation( 0 );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::InsertNodeList::InsertSteinerNotification( unsigned int const thread ) {

	if ( ! m_pPointData->IsErased() ) {

		RefcountedPointer< Operation > const& pLeftOperation  = m_pPointData->m_pWSP->m_pActiveOperations[ thread ];
		RefcountedPointer< Operation > const& pRightOperation = m_pPointData->m_pActiveInsertOperation;
		if ( ( pRightOperation.Get() != NULL ) && ( ( pLeftOperation.Get() == NULL ) || ( *pLeftOperation < *pRightOperation ) ) )
			m_pPointData->ScheduleInsertOperation( thread );
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::InsertNodeList::EraseNotification() {

	if ( m_pPointData->m_pActiveInsertOperation.Get() != NULL )
		m_pPointData->ScheduleInsertOperation( 0 );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::InsertNodeList::EraseSteinerNotification( unsigned int const thread ) {

	if ( ! m_pPointData->IsErased() ) {

		RefcountedPointer< Operation > const& pLeftOperation  = m_pPointData->m_pWSP->m_pActiveOperations[ thread ];
		RefcountedPointer< Operation > const& pRightOperation = m_pPointData->m_pActiveInsertOperation;
		if ( ( pRightOperation.Get() != NULL ) && ( ( pLeftOperation.Get() == NULL ) || ( *pLeftOperation < *pRightOperation ) ) )
			m_pPointData->ScheduleInsertOperation( thread );
	}
}




//============================================================================
//    WSP::PointData inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >::PointData::PointData(
	WSP* const pWSP,
	Handle const& handle,
	RefcountedPointer< Operation > const& pOperation
) :
	m_pWSP( pWSP ),
	m_handle( handle ),
	m_erased( false ),
	m_pOperation( pOperation ),
	m_initializeNodeList( this ),
	m_dispatchNodeList(   this ),
	m_insertNodeList(     this )
{
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::PointData::IsErased() const {

	return( __sync_bool_compare_and_swap( const_cast< bool volatile* >( &m_erased ), true, true ) );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::PointData::EraseSteiner( unsigned int const thread ) {

	BOOST_ASSERT( IsSteiner() );

	if ( __sync_bool_compare_and_swap( &m_erased, false, true ) ) {

		m_pWSP->m_erasedSteinerHandles.insert( m_handle );

		// recursively erase all child Steiner points
		{	typename std::set< RefcountedPointer< Operation > >::iterator ii    = m_pDispatchOperations.begin();
			typename std::set< RefcountedPointer< Operation > >::iterator iiEnd = m_pDispatchOperations.end();
			for ( ; ii != iiEnd; ++ii ) {

				BOOST_ASSERT( ( *ii )->GetSourceHandle() == m_handle );
				Handle const steinerHandle = ( *ii )->GetTargetHandle();
				BOOST_ASSERT( steinerHandle != m_handle );

				RefcountedPointer< PointData > const& pSteinerPointData( QT::GetClientData( steinerHandle ).Get() );
				BOOST_ASSERT( pSteinerPointData.Get() != NULL );
				BOOST_ASSERT( pSteinerPointData->GetHandle() == steinerHandle );
				BOOST_ASSERT( pSteinerPointData->IsSteiner() );

				BOOST_ASSERT( pSteinerPointData->m_pReverseDispatchOperation == *ii );
				pSteinerPointData->EraseSteiner( thread );
			}
		}

		m_pWSP->m_quadtree.EraseSteiner( m_handle, thread );
		{	typename std::set< Handle >::iterator ii = m_pWSP->m_recentlyInsertedSteinerHandles.find( m_handle );
			if ( ii != m_pWSP->m_recentlyInsertedSteinerHandles.end() )
				m_pWSP->m_recentlyInsertedSteinerHandles.erase( m_handle );
			else
				m_pWSP->m_recentlyErasedSteinerHandles.insert( m_handle );
		}
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::PointData::Cleanup( unsigned int const thread ) {

	m_pActiveInitializeOperation = RefcountedPointer< Operation >();
	m_pActiveDispatchOperation   = RefcountedPointer< Operation >();
	m_pActiveInsertOperation     = RefcountedPointer< Operation >();

	m_initializeNodeList.Clear();
	m_dispatchNodeList.Clear();
	m_insertNodeList.Clear();

	if ( m_pSelfInitializeOperation.Get() != NULL ) {

		BOOST_ASSERT( m_pSelfInitializeOperation->GetSourceHandle() == m_handle );
		BOOST_ASSERT( m_pSelfInitializeOperation->GetTargetHandle() == m_handle );

		m_pSelfInitializeOperation->SetEnqueued( false );

		m_pSelfInitializeOperation = RefcountedPointer< Operation >();
	}

	if ( m_pSelfDispatchOperation.Get() != NULL ) {

		BOOST_ASSERT( m_pReverseDispatchOperation == m_pSelfDispatchOperation );

		BOOST_ASSERT( m_pSelfDispatchOperation->GetSourceHandle() == m_handle );
		BOOST_ASSERT( m_pSelfDispatchOperation->GetTargetHandle() == m_handle );

		UndoDispatchOperation( thread );
		m_pSelfDispatchOperation->SetEnqueued( false );

		m_pSelfDispatchOperation = RefcountedPointer< Operation >();
		m_pReverseDispatchOperation = m_pSelfDispatchOperation;
	}

	if ( m_pSelfInsertOperation.Get() != NULL ) {

		BOOST_ASSERT( m_pReverseInsertOperations.find( m_pSelfInsertOperation ) != m_pReverseInsertOperations.end() );

		BOOST_ASSERT( m_pSelfInsertOperation->GetSourceHandle() == m_handle );
		BOOST_ASSERT( m_pSelfInsertOperation->GetTargetHandle() == m_handle );

		UndoInsertOperation( thread, m_pSelfInsertOperation );
		m_pSelfInsertOperation->SetEnqueued( false );
		BOOST_VERIFY( m_pReverseInsertOperations.erase( m_pSelfInsertOperation ) == 1 );

		m_pSelfInsertOperation = RefcountedPointer< Operation >();
	}

	{	typename std::set< RefcountedPointer< Operation > >::iterator ii    = m_pDispatchOperations.begin();
		typename std::set< RefcountedPointer< Operation > >::iterator iiEnd = m_pDispatchOperations.end();
		while ( ii != iiEnd ) {

			typename std::set< RefcountedPointer< Operation > >::iterator const iiCurrent = ii;
			++ii;

			BOOST_ASSERT( ( *iiCurrent )->GetSourceHandle() == m_handle );
			Handle const steinerHandle = ( *iiCurrent )->GetTargetHandle();
			BOOST_ASSERT( steinerHandle != m_handle );

			RefcountedPointer< PointData > const& pSteinerPointData( QT::GetClientData( steinerHandle ).Get() );
			BOOST_ASSERT( pSteinerPointData.Get() != NULL );
			BOOST_ASSERT( pSteinerPointData->GetHandle() == steinerHandle );
			BOOST_ASSERT( pSteinerPointData->IsSteiner() );

			BOOST_ASSERT( pSteinerPointData->m_pReverseDispatchOperation == *iiCurrent );
			( *iiCurrent )->SetEnqueued( false );

			{	RefcountedPointer< Operation > const pActiveOperation = m_pWSP->m_pActiveOperations[ thread ];
				m_pWSP->m_pActiveOperations[ thread ] = pSteinerPointData->m_pOperation;

				pSteinerPointData->EraseSteiner( thread );
				m_pWSP->m_erasedSteinerHandles.erase( steinerHandle );
				pSteinerPointData->Cleanup( thread );

				m_pWSP->m_pActiveOperations[ thread ] = pActiveOperation;
			}
		}
		BOOST_ASSERT( m_pDispatchOperations.empty() );
	}

	{	typename std::set< RefcountedPointer< Operation > >::iterator ii    = m_pInsertOperations.begin();
		typename std::set< RefcountedPointer< Operation > >::iterator iiEnd = m_pInsertOperations.end();
		for ( ; ii != iiEnd; ++ii ) {

			BOOST_ASSERT( ( *ii )->GetSourceHandle() == m_handle );
			Handle const targetHandle = ( *ii )->GetTargetHandle();

			RefcountedPointer< PointData > const& pTargetPointData( QT::GetClientData( targetHandle ).Get() );
			BOOST_ASSERT( pTargetPointData.Get() != NULL );
			BOOST_ASSERT( pTargetPointData->GetHandle() == targetHandle );

			pTargetPointData->UndoInsertOperation( thread, *ii );
			( *ii )->SetEnqueued( false );
			BOOST_VERIFY( pTargetPointData->m_pReverseInsertOperations.erase( *ii ) == 1 );
		}
		m_pInsertOperations.clear();
	}

	if ( m_pReverseDispatchOperation.Get() != NULL ) {

		BOOST_ASSERT( m_pReverseDispatchOperation->GetTargetHandle() == m_handle );
		Handle const sourceHandle = m_pReverseDispatchOperation->GetSourceHandle();

		RefcountedPointer< PointData > const& pSourcePointData( QT::GetClientData( sourceHandle ).Get() );
		BOOST_ASSERT( pSourcePointData.Get() != NULL );
		BOOST_ASSERT( pSourcePointData->GetHandle() == sourceHandle );

		m_pReverseDispatchOperation->SetEnqueued( false );
		BOOST_VERIFY( pSourcePointData->m_pDispatchOperations.erase( m_pReverseDispatchOperation ) == 1 );

		m_pReverseDispatchOperation = RefcountedPointer< Operation >();
	}

	{	typename std::set< RefcountedPointer< Operation > >::iterator ii    = m_pReverseInsertOperations.begin();
		typename std::set< RefcountedPointer< Operation > >::iterator iiEnd = m_pReverseInsertOperations.end();
		for ( ; ii != iiEnd; ++ii ) {

			BOOST_ASSERT( ( *ii )->GetTargetHandle() == m_handle );
			Handle const sourceHandle = ( *ii )->GetSourceHandle();

			RefcountedPointer< PointData > const& pSourcePointData( QT::GetClientData( sourceHandle ).Get() );
			BOOST_ASSERT( pSourcePointData.Get() != NULL );
			BOOST_ASSERT( pSourcePointData->GetHandle() == sourceHandle );

			( *ii )->SetEnqueued( false );
			BOOST_VERIFY( pSourcePointData->m_pInsertOperations.erase( *ii ) == 1 );
		}
		m_pReverseInsertOperations.clear();
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >* const WSP< t_Dimension, t_Threads, t_Type >::PointData::GetWSP() const {

	return m_pWSP;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
typename WSP< t_Dimension, t_Threads, t_Type >::Handle const& WSP< t_Dimension, t_Threads, t_Type >::PointData::GetHandle() const {

	return m_handle;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::PointData::IsSteiner() const {

	return( m_pOperation.Get() != NULL );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
RefcountedPointer< typename WSP< t_Dimension, t_Threads, t_Type >::Operation > const& WSP< t_Dimension, t_Threads, t_Type >::PointData::GetOperation() const {

	return m_pOperation;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::PointData::PerformInitializeOperation(
	PoolList& poolList,
	unsigned int const thread,
	RefcountedPointer< Operation > const& pOperation
)
{
	BOOST_ASSERT( ! IsErased() );
	BOOST_ASSERT( pOperation.Get() != NULL );
	BOOST_ASSERT( pOperation->GetSourceHandle() == m_handle );
	BOOST_ASSERT( pOperation->GetTargetHandle() == m_handle );

#ifndef NDEBUG
	{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

		BOOST_ASSERT( pOperation == m_pSelfInitializeOperation );

		BOOST_ASSERT( ! IsSteiner() );
		BOOST_ASSERT( m_initializeNodeList.IsEmpty() );
	}
#endif    // NDEBUG

	int const rank = m_pWSP->FindRank( Accuracy::ExactCast( m_pWSP->m_quadtree.LowerBoundNearestNeighborDistanceSquared( &m_initializeNodeList, m_handle ) ) );

	{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

		BOOST_ASSERT( m_pActiveInitializeOperation.Get() == NULL );
		m_pActiveInitializeOperation = pOperation;

		BOOST_ASSERT( m_pSelfDispatchOperation == m_pReverseDispatchOperation );

		if ( ( m_pSelfDispatchOperation.Get() == NULL ) || ( m_pSelfDispatchOperation->GetRank() != rank ) ) {

			if ( m_pSelfDispatchOperation.Get() != NULL ) {

				UndoDispatchOperation( thread );
				m_pSelfDispatchOperation->SetEnqueued( false );
			}

			m_pSelfDispatchOperation = RefcountedPointer< Operation >( new Operation( m_handle, m_handle, rank ) );
			m_pReverseDispatchOperation = m_pSelfDispatchOperation;
			if ( m_pActiveDispatchOperation.Get() == NULL ) {

				m_pWSP->m_pLocalDispatchOperations[ thread ].insert( m_pSelfDispatchOperation );
				m_pSelfDispatchOperation->SetEnqueued( true );
			}
			else
				ScheduleDispatchOperation( thread );
		}
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::PointData::PerformDispatchOperation(
	PoolList& poolList,
	unsigned int const thread,
	RefcountedPointer< Operation > const& pOperation
)
{
	BOOST_ASSERT( ! IsErased() );
	BOOST_ASSERT( pOperation.Get() != NULL );
	BOOST_ASSERT( pOperation->GetTargetHandle() == m_handle );

#ifndef NDEBUG
	{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

		BOOST_ASSERT( pOperation == m_pReverseDispatchOperation );

		BOOST_ASSERT( m_dispatchNodeList.IsEmpty() );
	}
#endif    // NDEBUG

	bool isSteiner = false;
	std::map< Handle, RefcountedPointer< Operation > > targetHandles;
	{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

		BOOST_ASSERT( m_pActiveDispatchOperation.Get() == NULL );
		m_pActiveDispatchOperation = pOperation;

		while ( ! m_pInsertOperations.empty() ) {

			RefcountedPointer< Operation > const pInsertOperation = *m_pInsertOperations.begin();

			BOOST_ASSERT( pInsertOperation->GetSourceHandle() == m_handle );
			Handle const targetHandle = pInsertOperation->GetTargetHandle();
			targetHandles.insert( std::pair< Handle, RefcountedPointer< Operation > >( targetHandle, pInsertOperation ) );

			RefcountedPointer< PointData > const& pTargetPointData( QT::GetClientData( targetHandle ).Get() );
			BOOST_ASSERT( pTargetPointData.Get() != NULL );
			BOOST_ASSERT( pTargetPointData->GetHandle() == targetHandle );

			BOOST_VERIFY( m_pInsertOperations.erase( pInsertOperation ) == 1 );
			BOOST_VERIFY( pTargetPointData->m_pReverseInsertOperations.erase( pInsertOperation ) == 1 );
		}

		isSteiner = IsSteiner();
	}

	boost::array< t_Type, t_Dimension > const& center( QT::GetPoint( m_handle ) );

	Functor const functor( m_pWSP, pOperation );

	Handle const nearestNeighborHandle = m_pWSP->m_quadtree.FindNearestNeighbor(
		&m_dispatchNodeList,
		m_handle,
		thread,
		functor
	);
	BOOST_ASSERT( nearestNeighborHandle != NULL );
	Accuracy::Exact const nearestNeighborDistanceSquared = ExactEuclideanDistanceSquared(
		center,
		QT::GetPoint( nearestNeighborHandle )
	);

	int const nearestNeighborRank = m_pWSP->FindRank( nearestNeighborDistanceSquared );

	if ( nearestNeighborRank >= pOperation->GetRank() ) {

		typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

		if ( ( m_pSelfInsertOperation.Get() == NULL ) || ( m_pSelfInsertOperation->GetRank() != nearestNeighborRank ) ) {

			if ( m_pSelfInsertOperation.Get() != NULL ) {

				UndoInsertOperation( thread, m_pSelfInsertOperation );
				m_pSelfInsertOperation->SetEnqueued( false );
				BOOST_VERIFY( m_pReverseInsertOperations.erase( m_pSelfInsertOperation ) == 1 );
			}

			m_pSelfInsertOperation = RefcountedPointer< Operation >( new Operation( m_handle, m_handle, nearestNeighborRank, true ) );
			m_pReverseInsertOperations.insert( m_pSelfInsertOperation );
			if ( m_pActiveInsertOperation.Get() == NULL ) {

				m_pWSP->m_pLocalInsertOperations[ thread ].insert( m_pSelfInsertOperation );
				m_pSelfInsertOperation->SetEnqueued( true );
			}
			else
				ScheduleInsertOperation( thread );
		}
	}

	if ( isSteiner ) {

		boost::shared_ptr< ClippedVoronoi< t_Dimension, t_Type, Handle > > pClippedVoronoi(
			m_pWSP->m_quadtree.FindClippedVoronoi(
				poolList,
				&m_dispatchNodeList,
				m_handle,
				nearestNeighborDistanceSquared,
				m_pWSP->m_rho,
				m_pWSP->m_beta,
				thread,
				functor
			)
		);

		unsigned int const iiEnd = pClippedVoronoi->GetNeighbors();
		for ( unsigned int ii = 0; ii < iiEnd; ++ii ) {

			ClientData< Handle > const* const pClientData = pClippedVoronoi->GetClientData( ii );
			if ( pClientData != NULL ) {    // ignore the "virtual" neighbors which define the bounding box

				Handle const neighborHandle = pClientData->Get();
				if ( neighborHandle != NULL ) {

					Accuracy::Exact const neighborDistanceSquared = ExactEuclideanDistanceSquared(
						center,
						QT::GetPoint( neighborHandle )
					);

					int const neighborRank = m_pWSP->FindRank( neighborDistanceSquared );

					if ( neighborRank >= pOperation->GetRank() ) {

						typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

						RefcountedPointer< PointData > const& pNeighborPointData( QT::GetClientData( neighborHandle ).Get() );
						BOOST_ASSERT( pNeighborPointData.Get() != NULL );
						BOOST_ASSERT( pNeighborPointData->GetHandle() == neighborHandle );

						typename std::map< Handle, RefcountedPointer< Operation > >::iterator const jj = targetHandles.find( m_handle );
						if ( ( jj == targetHandles.end() ) || ( jj->second->GetRank() != neighborRank ) ) {

							RefcountedPointer< Operation > pInsertOperation( new Operation( neighborHandle, m_handle, neighborRank, true ) );
							m_pInsertOperations.insert( pInsertOperation );
							pNeighborPointData->m_pReverseInsertOperations.insert( pInsertOperation );
							if ( pNeighborPointData->m_pActiveInsertOperation.Get() == NULL ) {

								m_pWSP->m_pLocalInsertOperations[ thread ].insert( pInsertOperation );
								pInsertOperation->SetEnqueued( true );
							}
							else
								pNeighborPointData->ScheduleInsertOperation( thread );
						}
						else {

							m_pInsertOperations.insert( jj->second );
							pNeighborPointData->m_pReverseInsertOperations.insert( jj->second );

							targetHandles.erase( jj );
						}
					}
				}
			}
		}
	}

	{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

		typename std::map< Handle, RefcountedPointer< Operation > >::const_iterator ii    = targetHandles.begin();
		typename std::map< Handle, RefcountedPointer< Operation > >::const_iterator iiEnd = targetHandles.end();
		for ( ; ii != iiEnd; ++ii ) {

			RefcountedPointer< PointData > const& pTargetPointData( QT::GetClientData( ii->first ).Get() );
			BOOST_ASSERT( pTargetPointData.Get() != NULL );
			BOOST_ASSERT( pTargetPointData->GetHandle() == ii->first );

			pTargetPointData->UndoInsertOperation( thread, ii->second );
			ii->second->SetEnqueued( false );
		}
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::PointData::PerformInsertOperation(
	PoolList& poolList,
	unsigned int const thread,
	RefcountedPointer< Operation > const& pOperation
)
{
	BOOST_ASSERT( ! IsErased() );
	BOOST_ASSERT( pOperation.Get() != NULL );
	BOOST_ASSERT( pOperation->GetTargetHandle() == m_handle );

	bool active = false;
	{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

		if ( m_pActiveInsertOperation.Get() == NULL ) {

			BOOST_ASSERT( m_insertNodeList.IsEmpty() );
			m_pActiveInsertOperation = pOperation;
			active = true;
		}
		else
			BOOST_ASSERT( *m_pActiveInsertOperation < *pOperation );
	}

	if ( active ) {

		boost::array< t_Type, t_Dimension > const& center( QT::GetPoint( m_handle ) );

		Functor const functor( m_pWSP, pOperation );

		Handle const nearestNeighborHandle = m_pWSP->m_quadtree.FindNearestNeighbor(
			&m_insertNodeList,
			m_handle,
			thread,
			functor
		);
		BOOST_ASSERT( nearestNeighborHandle != NULL );
		Accuracy::Exact const nearestNeighborDistanceSquared = ExactEuclideanDistanceSquared(
			center,
			QT::GetPoint( nearestNeighborHandle )
		);
		BOOST_ASSERT( pOperation->GetRank() == m_pWSP->FindRank( nearestNeighborDistanceSquared ) );

		boost::shared_ptr< ClippedVoronoi< t_Dimension, t_Type, Handle > > pClippedVoronoi(
			m_pWSP->m_quadtree.FindClippedVoronoi(
				poolList,
				&m_insertNodeList,
				m_handle,
				nearestNeighborDistanceSquared,
				m_pWSP->m_rho,
				m_pWSP->m_beta,
				thread,
				functor
			)
		);

		// see if any of the old Steiners are still valid
		{	typename std::vector< Handle > steinerHandles;
			{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

				typename std::set< RefcountedPointer< Operation > >::iterator ii    = m_pDispatchOperations.begin();
				typename std::set< RefcountedPointer< Operation > >::iterator iiEnd = m_pDispatchOperations.end();
				for ( ; ii != iiEnd; ++ii ) {

					BOOST_ASSERT( ( *ii )->GetSourceHandle() == m_handle );
					Handle const steinerHandle = ( *ii )->GetTargetHandle();
					BOOST_ASSERT( steinerHandle != m_handle );

					steinerHandles.push_back( steinerHandle );
				}
			}

			typename std::vector< Handle >::const_iterator ii    = steinerHandles.begin();
			typename std::vector< Handle >::const_iterator iiEnd = steinerHandles.end();
			for ( ; ii != iiEnd; ++ii ) {

				boost::array< t_Type, t_Dimension > const& steinerPoint( QT::GetPoint( *ii ) );
				bool const isSteiner = pClippedVoronoi->IsSteiner( steinerPoint );

				{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

					RefcountedPointer< PointData > const& pSteinerPointData( QT::GetClientData( *ii ).Get() );
					BOOST_ASSERT( pSteinerPointData.Get() != NULL );
					BOOST_ASSERT( pSteinerPointData->GetHandle() == *ii );
					BOOST_ASSERT( pSteinerPointData->IsSteiner() );

					BOOST_ASSERT( pSteinerPointData->m_pOperation.Get() != NULL );
					BOOST_ASSERT( *pOperation <= *pSteinerPointData->m_pOperation );

					if ( ( pSteinerPointData->m_pOperation == pOperation ) && isSteiner )
						pClippedVoronoi->InsertNeighbor( steinerPoint );
					else
						pSteinerPointData->EraseSteiner( thread );
				}
			}
		}

		for ( ; ; ) {

			boost::shared_ptr< boost::array< t_Type, t_Dimension > > const pSteinerPoint = pClippedVoronoi->FindSteiner( 0.9 );
			if ( pSteinerPoint == NULL )
				break;

			pClippedVoronoi->InsertNeighbor( *pSteinerPoint );

			Accuracy::Exact const steinerDistanceSquared = ExactEuclideanDistanceSquared( center, *pSteinerPoint );
			BOOST_ASSERT( steinerDistanceSquared >= nearestNeighborDistanceSquared * Square( Accuracy::ExactCast( m_pWSP->m_rho  ) ) );
			BOOST_ASSERT( steinerDistanceSquared <= nearestNeighborDistanceSquared * Square( Accuracy::ExactCast( m_pWSP->m_beta ) ) );

			int const steinerRank = m_pWSP->FindRank( steinerDistanceSquared );
			BOOST_ASSERT( steinerRank > pOperation->GetRank() );

			{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_pWSP->m_mutex );

				InsertCallback callback( m_pWSP, pOperation );
				Handle const steinerHandle = m_pWSP->m_quadtree.InsertSteiner( *pSteinerPoint, callback, thread );
				m_pWSP->m_recentlyInsertedSteinerHandles.insert( steinerHandle );
				BOOST_ASSERT( steinerHandle != NULL );

				RefcountedPointer< PointData > const& pSteinerPointData( QT::GetClientData( steinerHandle ).Get() );
				BOOST_ASSERT( pSteinerPointData.Get() != NULL );
				BOOST_ASSERT( pSteinerPointData->GetHandle() == steinerHandle );
				BOOST_ASSERT( pSteinerPointData->IsSteiner() );

				RefcountedPointer< Operation > const pDispatchOperation( new Operation( steinerHandle, m_handle, steinerRank ) );
				m_pDispatchOperations.insert( pDispatchOperation );
				pSteinerPointData->m_pReverseDispatchOperation = pDispatchOperation;

				BOOST_ASSERT( pSteinerPointData->m_pActiveDispatchOperation.Get() == NULL );
				m_pWSP->m_pLocalDispatchOperations[ thread ].insert( pDispatchOperation );
				pDispatchOperation->SetEnqueued( true );
			}
		}

		BOOST_ASSERT( pClippedVoronoi->IsWellSpaced() );
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::PointData::ScheduleInitializeOperation(
	unsigned int const thread,
	bool const force
)
{
	BOOST_ASSERT( m_pActiveInitializeOperation.Get() != NULL );

	if (
		force ||
		( ( m_pActiveInitializeOperation.Get() != NULL ) && ( ( m_pWSP->m_pActiveOperations[ thread ].Get() == NULL ) || ( *m_pWSP->m_pActiveOperations[ thread ] < *m_pActiveInitializeOperation ) ) )
	)
	{
		BOOST_ASSERT( m_pActiveInitializeOperation == m_pSelfInitializeOperation );
		BOOST_ASSERT( ! m_pSelfInitializeOperation->GetEnqueued() );

		m_pActiveInitializeOperation = RefcountedPointer< Operation >();
		m_initializeNodeList.Clear();

		m_pWSP->m_pInitializeOperations.insert( m_pSelfInitializeOperation );
		m_pSelfInitializeOperation->SetEnqueued( true );
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::PointData::ScheduleDispatchOperation(
	unsigned int const thread,
	bool const force
)
{
	BOOST_ASSERT( m_pActiveDispatchOperation.Get() != NULL );

	if (
		force ||
		( ( m_pActiveDispatchOperation.Get() != NULL ) && ( ( m_pWSP->m_pActiveOperations[ thread ].Get() == NULL ) || ( *m_pWSP->m_pActiveOperations[ thread ] < *m_pActiveDispatchOperation ) ) )
	)
	{
		BOOST_ASSERT( m_pActiveDispatchOperation == m_pReverseDispatchOperation );
		BOOST_ASSERT( ! m_pReverseDispatchOperation->GetEnqueued() );

		m_pActiveDispatchOperation = RefcountedPointer< Operation >();
		m_dispatchNodeList.Clear();

		m_pWSP->m_pLocalDispatchOperations[ thread ].insert( m_pReverseDispatchOperation );
		m_pReverseDispatchOperation->SetEnqueued( true );
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::PointData::ScheduleInsertOperation(
	unsigned int const thread,
	bool const force
)
{
	BOOST_ASSERT( m_pActiveInsertOperation.Get() != NULL );

	if (
		force ||
		( ( m_pActiveInsertOperation.Get() != NULL ) && ( ( m_pWSP->m_pActiveOperations[ thread ].Get() == NULL ) || ( *m_pWSP->m_pActiveOperations[ thread ] < *m_pActiveInsertOperation ) ) )
	)
	{
		m_pActiveInsertOperation = RefcountedPointer< Operation >();
		m_insertNodeList.Clear();

		typename std::set< RefcountedPointer< Operation > >::iterator ii    = m_pReverseInsertOperations.begin();
		typename std::set< RefcountedPointer< Operation > >::iterator iiEnd = m_pReverseInsertOperations.end();
		for ( ; ii != iiEnd; ++ii ) {

			if ( ( ii->Get() != NULL ) && ( ! ( *ii )->GetEnqueued() ) && ( ( m_pWSP->m_pActiveOperations[ thread ].Get() == NULL ) || ( *m_pWSP->m_pActiveOperations[ thread ] < **ii ) ) ) {

				m_pWSP->m_pLocalInsertOperations[ thread ].insert( *ii );
				( *ii )->SetEnqueued( true );
			}
		}
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::PointData::UndoDispatchOperation( unsigned int const thread ) {

	// only need to undo the self-insert operation, since the others cannot be scheduled at an incorrect rank

	if ( m_pSelfInsertOperation.Get() != NULL ) {

		BOOST_ASSERT( m_pReverseInsertOperations.find( m_pSelfInsertOperation ) != m_pReverseInsertOperations.end() );

		BOOST_ASSERT( m_pSelfInsertOperation->GetSourceHandle() == m_handle );
		BOOST_ASSERT( m_pSelfInsertOperation->GetTargetHandle() == m_handle );

		UndoInsertOperation( thread, m_pSelfInsertOperation );
		m_pSelfInsertOperation->SetEnqueued( false );
		BOOST_VERIFY( m_pReverseInsertOperations.erase( m_pSelfInsertOperation ) == 1 );

		m_pSelfInsertOperation = RefcountedPointer< Operation >();
	}

	if ( m_pActiveDispatchOperation.Get() != NULL ) {

		BOOST_ASSERT( m_pActiveDispatchOperation == m_pReverseDispatchOperation );
		ScheduleDispatchOperation( thread, true );
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::PointData::UndoInsertOperation(
	unsigned int const thread,
	RefcountedPointer< Operation > const& pInsertOperation
)
{
	typename std::set< RefcountedPointer< Operation > >::iterator ii    = m_pDispatchOperations.begin();
	typename std::set< RefcountedPointer< Operation > >::iterator iiEnd = m_pDispatchOperations.end();
	while ( ii != iiEnd ) {

		typename std::set< RefcountedPointer< Operation > >::iterator const iiCurrent = ii;
		++ii;

		BOOST_ASSERT( ( *iiCurrent )->GetSourceHandle() == m_handle );
		Handle const steinerHandle = ( *iiCurrent )->GetTargetHandle();
		BOOST_ASSERT( steinerHandle != m_handle );

		RefcountedPointer< PointData > const& pSteinerPointData( QT::GetClientData( steinerHandle ).Get() );
		BOOST_ASSERT( pSteinerPointData.Get() != NULL );
		BOOST_ASSERT( pSteinerPointData->GetHandle() == steinerHandle );
		BOOST_ASSERT( pSteinerPointData->IsSteiner() );

		if ( pSteinerPointData->m_pOperation == pInsertOperation ) {

			( *iiCurrent )->SetEnqueued( false );
			BOOST_ASSERT( pSteinerPointData->m_pReverseDispatchOperation == *iiCurrent );
			pSteinerPointData->EraseSteiner( thread );
		}
	}

	if ( m_pActiveInsertOperation == pInsertOperation )
		ScheduleInsertOperation( thread, true );
}




//============================================================================
//    WSP inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
WSP< t_Dimension, t_Threads, t_Type >::WSP(
	double const rho,
	double const beta,
	unsigned int const maximumColors,
	unsigned int const nodeSize,
	double const delta
) :
	m_rho( rho ),
	m_logRankRho( 0.9 * std::log( m_rho ) ),
	m_rankRho( std::exp( m_logRankRho ) ),
	m_beta( beta ),
	m_rank( std::numeric_limits< int >::max() + 1 ),
	m_color( 0 ),
	m_quadtree( nodeSize, delta )
{
	BOOST_ASSERT( maximumColors >= 2 );

	/*
		For the m_colors calculation:
			double const denominator = std::pow( m_rankRho, m_rank - epsilon ) / std::sqrt( t_Dimension );
			double const distance = 3.0 * m_beta * std::pow( m_rankRho, m_rank + 1 + epsilon );
			double const ratio = distance / denominator;
			unsigned int const colors = std::ceil( ratio ) + 1;
	*/

	double const ratio = 3.0 * m_beta * std::pow( m_rankRho, 1.2 ) * std::sqrt( t_Dimension );    // the powers of m_rank cancel

	m_colorScale = std::max( 1.0, ratio / ( maximumColors - 1.1 ) );
	m_colors = std::ceil( ratio / m_colorScale ) + 1;
	BOOST_ASSERT( m_colors <= maximumColors );

	if ( m_rho <= 1 )
		throw std::runtime_error( "Rho must be at least 1" );
	if ( ( m_beta <= m_rho ) || ( m_beta >= 2 * m_rho ) )
		throw std::runtime_error( "Beta must be between rho and 2*rho" );
	if ( ( HighBit( m_colors ) + 1 ) * t_Dimension > 8 * sizeof( uint64_t ) - 1 )    // typeof( m_color ) = uint64_t, and we reserve one bit
		throw std::runtime_error( "Too many colors are required" );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
boost::array< t_Type, t_Dimension > const& WSP< t_Dimension, t_Threads, t_Type >::GetPoint( Handle const& handle ) {

	BOOST_ASSERT( handle != NULL );
	return QT::GetPoint( handle );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::IsSteiner( Handle const& handle ) {

	RefcountedPointer< PointData > const& pPointData( QT::GetClientData( handle ).Get() );
	return pPointData->IsSteiner();
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
unsigned int const WSP< t_Dimension, t_Threads, t_Type >::GetPoints() const {

	return m_quadtree.GetPoints();
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
unsigned int const WSP< t_Dimension, t_Threads, t_Type >::GetSteinerPoints() const {

	return m_quadtree.GetSteinerPoints();
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::EnumeratePoints( boost::function1< bool const, Handle const& > const& enumerator ) const {

	return m_quadtree.EnumeratePoints( enumerator );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::EnumerateSteinerPoints( boost::function1< bool const, Handle const& > const& enumerator ) const {

	return m_quadtree.EnumerateSteinerPoints( enumerator );
}

template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::EnumerateRecentlyInsertedSteinerPoints( boost::function1< bool const, Handle const& > const& enumerator ) const {

	bool result = true;

	typename std::set< Handle >::const_iterator ii    = m_lessRecentlyInsertedSteinerHandles.begin();
	typename std::set< Handle >::const_iterator iiEnd = m_lessRecentlyInsertedSteinerHandles.end();
	for ( ; result && ( ii != iiEnd ); ++ii )
		result = enumerator( *ii );

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::EnumerateRecentlyErasedSteinerPoints( boost::function1< bool const, Handle const& > const& enumerator ) const {

	bool result = true;

	typename std::set< Handle >::const_iterator ii    = m_lessRecentlyErasedSteinerHandles.begin();
	typename std::set< Handle >::const_iterator iiEnd = m_lessRecentlyErasedSteinerHandles.end();
	for ( ; result && ( ii != iiEnd ); ++ii )
		result = enumerator( *ii );

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
typename WSP< t_Dimension, t_Threads, t_Type >::Handle const WSP< t_Dimension, t_Threads, t_Type >::FindNearestNeighbor( Handle const& handle ) const {

	return m_quadtree.FindNearestNeighbor( NULL, handle );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
bool const WSP< t_Dimension, t_Threads, t_Type >::EnumerateVoronoiNeighbors(
	Handle const& handle,
	boost::function1< bool const, Handle const& > const& enumerator
) const
{
#ifndef NDEBUG
	for ( unsigned int ii = 0; ii < t_Threads; ++ii ) {

		BOOST_ASSERT( m_pLocalDispatchOperations[ ii ].empty() );
		BOOST_ASSERT( m_pLocalInsertOperations[   ii ].empty() );
	}
#endif    // NDEBUG

	if ( ! ( m_pInitializeOperations.empty() && m_pDispatchOperations.empty() && m_pInsertOperations.empty() ) )
		throw std::runtime_error( "Voronoi neighbors are only available for a well-spaced point set" );

	Handle const nearestNeighborHandle = m_quadtree.FindNearestNeighbor( NULL, handle );
	BOOST_ASSERT( nearestNeighborHandle != NULL );
	Accuracy::Exact const nearestNeighborDistanceSquared = ExactEuclideanDistanceSquared(
		QT::GetPoint( handle ),
		QT::GetPoint( nearestNeighborHandle )
	);

	PoolList poolList;
	boost::shared_ptr< ClippedVoronoi< t_Dimension, t_Type, Handle > > pClippedVoronoi(
		m_quadtree.FindClippedVoronoi(
			poolList,
			NULL,
			handle,
			nearestNeighborDistanceSquared,
			m_rho,
			m_rho
		)
	);

	bool result = true;
	unsigned int const iiEnd = pClippedVoronoi->GetNeighbors();
	for ( unsigned int ii = 0; result && ( ii < iiEnd ); ++ii ) {

		ClientData< Handle > const* const pClientData = pClippedVoronoi->GetClientData( ii );
		if ( pClientData != NULL )    // ignore the "virtual" neighbors which define the bounding box
			result = enumerator( pClientData->Get() );
	}

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
typename WSP< t_Dimension, t_Threads, t_Type >::Handle const WSP< t_Dimension, t_Threads, t_Type >::Insert( boost::array< t_Type, t_Dimension > const& vertex ) {

	for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

		if ( ( vertex[ ii ] <= 0 ) || ( vertex[ ii ] >= 1 ) )
			throw std::runtime_error( "Each coordinate must be in (0,1)" );
	}

	InsertCallback callback( this, RefcountedPointer< Operation >() );
	Handle const handle = m_quadtree.Insert( vertex, callback );
	if ( handle != NULL ) {

		RefcountedPointer< PointData > const& pPointData( QT::GetClientData( handle ).Get() );
		BOOST_ASSERT( pPointData.Get() != NULL );
		BOOST_ASSERT( pPointData->GetHandle() == handle );
		BOOST_ASSERT( ! pPointData->IsSteiner() );

		RefcountedPointer< Operation > const pInitializeOperation( new Operation( handle, handle ) );
		pPointData->m_pSelfInitializeOperation = pInitializeOperation;
		m_pInitializeOperations.insert( pInitializeOperation );
		pInitializeOperation->SetEnqueued( true );
	}

	// erase any Steiner points that need erasing
	while ( ! m_erasedSteinerHandles.empty() ) {

		Handle const steinerHandle = *m_erasedSteinerHandles.begin();
		m_erasedSteinerHandles.erase( m_erasedSteinerHandles.begin() );

		RefcountedPointer< PointData > const& pSteinerPointData( QT::GetClientData( steinerHandle ).Get() );
		BOOST_ASSERT( pSteinerPointData.Get() != NULL );
		BOOST_ASSERT( pSteinerPointData->GetHandle() == steinerHandle );
		BOOST_ASSERT( pSteinerPointData->IsSteiner() );

		m_pActiveOperations[ 0 ] = pSteinerPointData->m_pOperation;
		pSteinerPointData->Cleanup( 0 );
		m_pActiveOperations[ 0 ] = RefcountedPointer< Operation >();
	}
	m_quadtree.CleanupErased();

	// move enqueued operations from the local into the global queues
	m_pDispatchOperations.insert( m_pLocalDispatchOperations[ 0 ].begin(), m_pLocalDispatchOperations[ 0 ].end() );
	m_pInsertOperations.insert(   m_pLocalInsertOperations[   0 ].begin(), m_pLocalInsertOperations[   0 ].end() );
	m_pLocalDispatchOperations[ 0 ].clear();
	m_pLocalInsertOperations[   0 ].clear();

	return handle;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::Erase( Handle const& handle ) {

	RefcountedPointer< PointData > const& pPointData( QT::GetClientData( handle ).Get() );
	BOOST_ASSERT( pPointData.Get() != NULL );
	BOOST_ASSERT( pPointData->GetHandle() == handle );

	if ( pPointData->IsSteiner() )
		throw std::runtime_error( "You cannot erase a Steiner point" );

	pPointData->Cleanup( 0 );
	m_quadtree.Erase( handle );

	// erase any Steiner points that need erasing
	while ( ! m_erasedSteinerHandles.empty() ) {

		Handle const steinerHandle = *m_erasedSteinerHandles.begin();
		m_erasedSteinerHandles.erase( m_erasedSteinerHandles.begin() );

		RefcountedPointer< PointData > const& pSteinerPointData( QT::GetClientData( steinerHandle ).Get() );
		BOOST_ASSERT( pSteinerPointData.Get() != NULL );
		BOOST_ASSERT( pSteinerPointData->GetHandle() == steinerHandle );
		BOOST_ASSERT( pSteinerPointData->IsSteiner() );

		m_pActiveOperations[ 0 ] = pSteinerPointData->m_pOperation;
		pSteinerPointData->Cleanup( 0 );
		m_pActiveOperations[ 0 ] = RefcountedPointer< Operation >();
	}
	m_quadtree.CleanupErased();

	// move enqueued operations from the local into the global queues
	m_pDispatchOperations.insert( m_pLocalDispatchOperations[ 0 ].begin(), m_pLocalDispatchOperations[ 0 ].end() );
	m_pInsertOperations.insert(   m_pLocalInsertOperations[   0 ].begin(), m_pLocalInsertOperations[   0 ].end() );
	m_pLocalDispatchOperations[ 0 ].clear();
	m_pLocalInsertOperations[   0 ].clear();
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
int const WSP< t_Dimension, t_Threads, t_Type >::FindRank( Accuracy::Exact const& distanceSquared ) const {

	BOOST_ASSERT( m_logRankRho > 0 );
	int rank = std::floor( 0.5 * std::log( distanceSquared.GetApproximate().GetValue() ) / m_logRankRho );

	if ( Accuracy::ExactCast( std::pow( m_rankRho, 2 * rank ) ) > distanceSquared ) {

		--rank;
		while ( Accuracy::ExactCast( std::pow( m_rankRho, 2 * rank ) ) > distanceSquared )
			--rank;
	}
	else {

		while ( Accuracy::ExactCast( std::pow( m_rankRho, 2 * ( rank + 1 ) ) ) <= distanceSquared )
			++rank;
	}
	BOOST_ASSERT( Accuracy::ExactCast( std::pow( m_rankRho, 2 *   rank       ) ) <= distanceSquared );
	BOOST_ASSERT( Accuracy::ExactCast( std::pow( m_rankRho, 2 * ( rank + 1 ) ) ) >  distanceSquared );

	return rank;
}




//============================================================================
//    WSP methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::Refine( unsigned int const threads ) {

	if ( threads < 1 )
		throw std::runtime_error( "Must use at least one thread" );
	if ( threads > t_Threads )
		throw std::runtime_error( "Requested more threads than the maximum" );

	BOOST_ASSERT( m_rank == std::numeric_limits< int >::max() + 1 );
	BOOST_ASSERT( m_color == 0 );

	m_lessRecentlyInsertedSteinerHandles.clear();
	m_lessRecentlyErasedSteinerHandles.clear();

	typename MaybeThreaded< ( t_Threads > 1 ) >::Barrier barrier( threads );
	if ( threads == 1 )
		Thread( &barrier, 0 );
	else {

		boost::thread_group threadGroup;
		for ( unsigned int ii = 0; ii < threads; ++ii ) {

			threadGroup.create_thread(
				boost::bind(
					boost::mem_fn( &WSP::Thread ),
					this,
					&barrier,
					ii
				)
			);
		}
		threadGroup.join_all();
	}

	m_rank = std::numeric_limits< int >::max() + 1;
	m_color = 0;

	m_lessRecentlyInsertedSteinerHandles.swap( m_recentlyInsertedSteinerHandles );
	m_lessRecentlyErasedSteinerHandles.swap(   m_recentlyErasedSteinerHandles   );

	BOOST_ASSERT( m_pInitializeOperations.empty() );
	BOOST_ASSERT( m_pDispatchOperations.empty()   );
	BOOST_ASSERT( m_pInsertOperations.empty()     );

#ifndef NDEBUG
	for ( unsigned int ii = 0; ii < t_Threads; ++ii ) {

		BOOST_ASSERT( m_pLocalDispatchOperations[ ii ].empty() );
		BOOST_ASSERT( m_pLocalInsertOperations[   ii ].empty() );
	}
#endif    // NDEBUG
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type >
void WSP< t_Dimension, t_Threads, t_Type >::Thread(
	typename MaybeThreaded< ( t_Threads > 1 ) >::Barrier* const pBarrier,
	unsigned int const thread
)
{
	PoolList poolList;

	std::multiset< RefcountedPointer< Operation >, CompareOperations > pOperations;

	pBarrier->wait();

	// perform initialization operations
	for ( ; ; ) {

		BOOST_ASSERT( pOperations.empty() );
		{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_mutex );

			typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator const iiBegin = m_pInitializeOperations.begin();
			typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator const iiEnd   = m_pInitializeOperations.end();
			if ( iiBegin == iiEnd )
				break;
			RefcountedPointer< Operation > const& pOperation = *iiBegin;
			typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator ii = iiBegin;
			for ( ++ii; ( ii != iiEnd ) && ( ( *ii )->GetTargetHandle() == pOperation->GetTargetHandle() ); ++ii );    // **NOTE: duplicate operations
			pOperations.insert( iiBegin, ii );
			m_pInitializeOperations.erase( iiBegin, ii );
		}
		BOOST_ASSERT( ! pOperations.empty() );

		while ( ! pOperations.empty() ) {

			typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::const_iterator ii = pOperations.begin();
			RefcountedPointer< Operation > const pOperation = *ii;
			pOperations.erase( ii );

			BOOST_ASSERT( pOperation->GetRank()            == std::numeric_limits< int >::max() + 1 );
			BOOST_ASSERT( pOperation->GetColor()           == 0                                     );
			BOOST_ASSERT( pOperation->GetColorCoordinate() == 0                                     );

			// begin initialization

			if ( pOperation->GetEnqueued() ) {

				BOOST_ASSERT( pOperation->GetSourceHandle() == pOperation->GetTargetHandle() );

				RefcountedPointer< PointData > const& pTargetPointData( QT::GetClientData( pOperation->GetTargetHandle() ).Get() );
				BOOST_ASSERT( pTargetPointData.Get() != NULL );
				BOOST_ASSERT( pTargetPointData->GetHandle() == pOperation->GetTargetHandle() );

				m_pActiveOperations[ thread ] = pOperation;
				pTargetPointData->PerformInitializeOperation( poolList, thread, pOperation );
				m_pActiveOperations[ thread ] = RefcountedPointer< Operation >();

				pOperation->SetEnqueued( false );
			}

			// end initialization
		}
		BOOST_ASSERT( pOperations.empty() );
	}

	// move enqueued operations from the local into the global queues
	{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_mutex );

		m_pDispatchOperations.insert( m_pLocalDispatchOperations[ thread ].begin(), m_pLocalDispatchOperations[ thread ].end() );
		m_pInsertOperations.insert(   m_pLocalInsertOperations[   thread ].begin(), m_pLocalInsertOperations[   thread ].end() );
		m_pLocalDispatchOperations[ thread ].clear();
		m_pLocalInsertOperations[   thread ].clear();
	}

	for ( ; ; ) {

		pBarrier->wait();

		// find the new rank at which we're working
		if ( thread == 0 ) {

			int rank = std::numeric_limits< int >::max();
			{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_mutex );

				if ( ! m_pInitializeOperations.empty() )
					rank = std::min( rank, ( *m_pInitializeOperations.begin() )->GetRank() );
				if ( ! m_pDispatchOperations.empty() )
					rank = std::min( rank, ( *m_pDispatchOperations.begin() )->GetRank() );
				if ( ! m_pInsertOperations.empty() )
					rank = std::min( rank, ( *m_pInsertOperations.begin() )->GetRank() );
			}

			m_rank = rank;
		}

		pBarrier->wait();

		if ( m_rank == std::numeric_limits< int >::max() )
			break;
		BOOST_ASSERT( m_rank <= FindRank( Accuracy::ExactCast( t_Dimension ) ) );

		// perform dispatch operations
		for ( ; ; ) {

			BOOST_ASSERT( pOperations.empty() );
			{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_mutex );

				typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator const iiBegin = m_pDispatchOperations.begin();
				typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator const iiEnd   = m_pDispatchOperations.end();
				if ( iiBegin == iiEnd )
					break;
				RefcountedPointer< Operation > const& pOperation = *iiBegin;
				if ( pOperation->GetRank() != m_rank )
					break;
				typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator ii = iiBegin;
				for ( ++ii; ( ii != iiEnd ) && ( ( *ii )->GetRank() == m_rank ) && ( ( *ii )->GetTargetHandle() == pOperation->GetTargetHandle() ); ++ii );    // **NOTE: duplicate operations
				pOperations.insert( iiBegin, ii );
				m_pDispatchOperations.erase( iiBegin, ii );
			}
			BOOST_ASSERT( ! pOperations.empty() );

			while ( ! pOperations.empty() ) {

				typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::const_iterator ii = pOperations.begin();
				RefcountedPointer< Operation > const pOperation = *ii;
				pOperations.erase( ii );

				BOOST_ASSERT( pOperation->GetRank()            == m_rank );
				BOOST_ASSERT( pOperation->GetColor()           == 0      );
				BOOST_ASSERT( pOperation->GetColorCoordinate() == 0      );

				// begin dispatch

				if ( pOperation->GetEnqueued() ) {

					RefcountedPointer< PointData > const& pTargetPointData( QT::GetClientData( pOperation->GetTargetHandle() ).Get() );
					BOOST_ASSERT( pTargetPointData.Get() != NULL );
					BOOST_ASSERT( pTargetPointData->GetHandle() == pOperation->GetTargetHandle() );

					m_pActiveOperations[ thread ] = pOperation;
					pTargetPointData->PerformDispatchOperation( poolList, thread, pOperation );
					m_pActiveOperations[ thread ] = RefcountedPointer< Operation >();

					pOperation->SetEnqueued( false );
				}

				// end dispatch
			}
			BOOST_ASSERT( pOperations.empty() );
		}

		pBarrier->wait();

		if ( thread == 0 ) {

			// erase any Steiner points that need erasing
			while ( ! m_erasedSteinerHandles.empty() ) {

				Handle const steinerHandle = *m_erasedSteinerHandles.begin();
				m_erasedSteinerHandles.erase( m_erasedSteinerHandles.begin() );

				RefcountedPointer< PointData > const& pSteinerPointData( QT::GetClientData( steinerHandle ).Get() );
				BOOST_ASSERT( pSteinerPointData.Get() != NULL );
				BOOST_ASSERT( pSteinerPointData->GetHandle() == steinerHandle );
				BOOST_ASSERT( pSteinerPointData->IsSteiner() );

				m_pActiveOperations[ thread ] = pSteinerPointData->m_pOperation;
				pSteinerPointData->Cleanup( thread );
				m_pActiveOperations[ thread ] = RefcountedPointer< Operation >();
			}
			m_quadtree.CleanupErased();
		}

		pBarrier->wait();

		// move enqueued operations from the local into the global queues
		{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_mutex );

			m_pDispatchOperations.insert( m_pLocalDispatchOperations[ thread ].begin(), m_pLocalDispatchOperations[ thread ].end() );
			m_pInsertOperations.insert(   m_pLocalInsertOperations[   thread ].begin(), m_pLocalInsertOperations[   thread ].end() );
			m_pLocalDispatchOperations[ thread ].clear();
			m_pLocalInsertOperations[   thread ].clear();
		}

		for ( ; ; ) {

			pBarrier->wait();

			// find the new color at which we're working
			if ( thread == 0 ) {

				uint64_t color = 0;

				typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator const iiBegin = m_pInsertOperations.begin();
				typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator const iiEnd   = m_pInsertOperations.end();
				if ( iiBegin != iiEnd ) {

					RefcountedPointer< Operation > const& pOperation = *iiBegin;
					if ( pOperation->GetRank() == m_rank )
						color = pOperation->GetColor();
				}

				m_color = color;
			}

			pBarrier->wait();

			if ( m_color == 0 )
				break;

			// perform Steiners insertion operations
			for ( ; ; ) {

				BOOST_ASSERT( pOperations.empty() );
				uint64_t colorCoordinate = 0;
				{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_mutex );

					typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator const iiBegin = m_pInsertOperations.begin();
					typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator const iiEnd   = m_pInsertOperations.end();
					if ( iiBegin == iiEnd )
						break;
					RefcountedPointer< Operation > const& pOperation = *iiBegin;
					if ( ( pOperation->GetRank() != m_rank ) || ( pOperation->GetColor() != m_color ) )
						break;
					colorCoordinate = pOperation->GetColorCoordinate();
					typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::iterator ii = iiBegin;
					for ( ++ii; ( ii != iiEnd ) && ( ( *ii )->GetRank() == m_rank ) && ( ( *ii )->GetColor() == m_color ) && ( ( *ii )->GetColorCoordinate() == colorCoordinate ); ++ii );
					pOperations.insert( iiBegin, ii );
					m_pInsertOperations.erase( iiBegin, ii );
				}
				BOOST_ASSERT( ! pOperations.empty() );

				while ( ! pOperations.empty() ) {

					typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::const_iterator ii = pOperations.begin();
					RefcountedPointer< Operation > const pOperation = *ii;
					pOperations.erase( ii );

					BOOST_ASSERT( pOperation->GetRank()  == m_rank  );
					BOOST_ASSERT( pOperation->GetColor() == m_color );

					// begin insert

					if ( pOperation->GetEnqueued() ) {

						RefcountedPointer< PointData > const& pTargetPointData( QT::GetClientData( pOperation->GetTargetHandle() ).Get() );
						BOOST_ASSERT( pTargetPointData.Get() != NULL );
						BOOST_ASSERT( pTargetPointData->GetHandle() == pOperation->GetTargetHandle() );

						m_pActiveOperations[ thread ] = pOperation;
						pTargetPointData->PerformInsertOperation( poolList, thread, pOperation );
						m_pActiveOperations[ thread ] = RefcountedPointer< Operation >();

						pOperation->SetEnqueued( false );
					}

					// end insert

					// any newly-queued insert operations at this rank,color,color-coordinate should be executed immediately
					if ( ! m_pLocalInsertOperations[ thread ].empty() ) {

						typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_mutex );

						typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::const_iterator ii    = m_pLocalInsertOperations[ thread ].begin();
						typename std::multiset< RefcountedPointer< Operation >, CompareOperations >::const_iterator iiEnd = m_pLocalInsertOperations[ thread ].end();
						for ( ; ii != iiEnd; ++ii ) {

							if ( ( ( *ii )->GetRank() == m_rank ) && ( ( *ii )->GetColor() == m_color ) && ( ( *ii )->GetColorCoordinate() == colorCoordinate ) )
								pOperations.insert( *ii );
							else
								m_pInsertOperations.insert( *ii );
						}

						m_pLocalInsertOperations[ thread ].clear();
					}
				}
				BOOST_ASSERT( pOperations.empty() );
			}

			pBarrier->wait();

			if ( thread == 0 ) {

				// erase any Steiner points that need erasing
				while ( ! m_erasedSteinerHandles.empty() ) {

					Handle const steinerHandle = *m_erasedSteinerHandles.begin();
					m_erasedSteinerHandles.erase( m_erasedSteinerHandles.begin() );

					RefcountedPointer< PointData > const& pSteinerPointData( QT::GetClientData( steinerHandle ).Get() );
					BOOST_ASSERT( pSteinerPointData.Get() != NULL );
					BOOST_ASSERT( pSteinerPointData->GetHandle() == steinerHandle );
					BOOST_ASSERT( pSteinerPointData->IsSteiner() );

					m_pActiveOperations[ thread ] = pSteinerPointData->m_pOperation;
					pSteinerPointData->Cleanup( thread );
					m_pActiveOperations[ thread ] = RefcountedPointer< Operation >();
				}
				m_quadtree.CleanupErased();
			}

			pBarrier->wait();

			// move enqueued operations from the local into the global queues
			{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_mutex );

				m_pDispatchOperations.insert( m_pLocalDispatchOperations[ thread ].begin(), m_pLocalDispatchOperations[ thread ].end() );
				m_pInsertOperations.insert(   m_pLocalInsertOperations[   thread ].begin(), m_pLocalInsertOperations[   thread ].end() );
				m_pLocalDispatchOperations[ thread ].clear();
				m_pLocalInsertOperations[   thread ].clear();
			}
		}

#ifndef NDEBUG
		{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock lock( m_mutex );

			// make sure that no operations are scheduled before the current rank
			BOOST_ASSERT( m_pInitializeOperations.empty() || ( ( *m_pInitializeOperations.begin() )->GetRank() > m_rank ) );
			BOOST_ASSERT( m_pDispatchOperations.empty()   || ( ( *m_pDispatchOperations.begin()   )->GetRank() > m_rank ) );
			BOOST_ASSERT( m_pInsertOperations.empty()     || ( ( *m_pInsertOperations.begin()     )->GetRank() > m_rank ) );
		}
#endif    // NDEBUG
	}

	BOOST_ASSERT( m_pActiveOperations[ thread ].Get() == NULL );
}




#endif    /* __cplusplus */

#endif    /* __WSP_HH__ */
